"""Tests for the Home related functionality."""

import base64
from collections.abc import Iterator
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from roborock.data.containers import CombinedMapInfo, HomeDataRoom, NamedRoomMapping
from roborock.data.v1.v1_code_mappings import RoborockStateCode
from roborock.data.v1.v1_containers import MultiMapsListMapInfo, MultiMapsListRoom
from roborock.devices.cache import DeviceCache, DeviceCacheData, InMemoryCache
from roborock.devices.device import RoborockDevice
from roborock.devices.traits.v1.home import HomeTrait
from roborock.devices.traits.v1.map_content import MapContentTrait
from roborock.devices.traits.v1.maps import MapsTrait
from roborock.devices.traits.v1.rooms import RoomsTrait
from roborock.devices.traits.v1.status import StatusTrait
from roborock.exceptions import RoborockDeviceBusy, RoborockException, RoborockInvalidStatus
from roborock.map.map_parser import ParsedMapData
from roborock.roborock_typing import RoborockCommand
from tests import mock_data

MULTI_MAP_LIST_DATA = [
    {
        "max_multi_map": 2,
        "max_bak_map": 1,
        "multi_map_count": 2,
        "map_info": [
            {
                "mapFlag": 0,
                "add_time": 1747132930,
                "length": 0,
                "name": "Ground Floor",
                "bak_maps": [{"mapFlag": 4, "add_time": 1747132936}],
            },
            {
                "mapFlag": 123,
                "add_time": 1747132940,
                "length": 0,
                "name": "Second Floor",
                "bak_maps": [{"mapFlag": 5, "add_time": 1747132946}],
            },
        ],
    }
]
MULTI_MAP_LIST_SINGLE_MAP_DATA = [
    {
        "max_multi_map": 1,
        "max_bak_map": 0,
        "multi_map_count": 1,
        "map_info": [
            {
                "mapFlag": 0,
                "add_time": 1747132930,
                "length": 0,
                "name": "Only Floor",
                "bak_maps": [],
            },
        ],
    }
]

ROOM_MAPPING_DATA_MAP_0 = [[16, "2362048"], [17, "2362044"]]
ROOM_MAPPING_DATA_MAP_123 = [[18, "2362041"], [19, "2362042"]]

UPDATED_STATUS_MAP_0 = {
    **mock_data.STATUS,
    "map_status": 0 * 4 + 3,  # Set current map to 0
}

UPDATED_STATUS_MAP_123 = {
    **mock_data.STATUS,
    "map_status": 123 * 4 + 3,  # Set current map to 123
}
MAP_BYTES_RESPONSE_1 = b"<map bytes 1>"
MAP_BYTES_RESPONSE_2 = b"<map bytes 2>"
TEST_IMAGE_CONTENT_1 = b"<image bytes 1>"
TEST_IMAGE_CONTENT_2 = b"<image bytes 2>"
TEST_PARSER_MAP = {
    MAP_BYTES_RESPONSE_1: TEST_IMAGE_CONTENT_1,
    MAP_BYTES_RESPONSE_2: TEST_IMAGE_CONTENT_2,
}


@pytest.fixture(autouse=True)
def no_sleep() -> Iterator[None]:
    """Patch sleep to avoid delays in tests."""
    with patch("roborock.devices.traits.v1.home.asyncio.sleep"):
        yield


@pytest.fixture(name="cache")
def cache_fixture():
    """Create an in-memory cache for testing."""
    return InMemoryCache()


@pytest.fixture(name="device_cache")
def device_cache_fixture(cache: InMemoryCache) -> DeviceCache:
    """Create a DeviceCache instance for testing."""
    return DeviceCache(duid="abc123", cache=cache)


@pytest.fixture(autouse=True)
async def status_trait(mock_rpc_channel: AsyncMock, device: RoborockDevice) -> StatusTrait:
    """Create a StatusTrait instance with mocked dependencies."""
    assert device.v1_properties
    status_trait = device.v1_properties.status

    # Verify initial state
    assert status_trait.current_map is None
    mock_rpc_channel.send_command.side_effect = [UPDATED_STATUS_MAP_0]
    await status_trait.refresh()
    assert status_trait.current_map == 0

    mock_rpc_channel.reset_mock()
    return status_trait


@pytest.fixture
def maps_trait(device: RoborockDevice) -> MapsTrait:
    """Create a MapsTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.maps


@pytest.fixture
def map_content_trait(device: RoborockDevice) -> MapContentTrait:
    """Create a MapContentTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.map_content


@pytest.fixture
def rooms_trait(device: RoborockDevice) -> RoomsTrait:
    """Create a RoomsTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.rooms


@pytest.fixture
def mock_web_api(web_api_client: AsyncMock) -> AsyncMock:
    """Alias the shared web API fixture for readability in this module."""
    web_api_client.get_rooms.return_value = []
    return web_api_client


@pytest.fixture
def home_trait(
    status_trait: StatusTrait,
    maps_trait: MapsTrait,
    map_content_trait: MapContentTrait,
    rooms_trait: RoomsTrait,
    device_cache: DeviceCache,
) -> HomeTrait:
    """Create a HomeTrait instance with mocked dependencies."""
    return HomeTrait(status_trait, maps_trait, map_content_trait, rooms_trait, device_cache)


@pytest.fixture(autouse=True)
def map_parser_fixture() -> Iterator[None]:
    """Mock MapParser.parse to return predefined test map data."""

    def parse_data(response: bytes) -> ParsedMapData:
        if image_content := TEST_PARSER_MAP.get(response):
            return ParsedMapData(
                image_content=image_content,
                map_data=MagicMock(),
            )
        raise ValueError(f"Unexpected map bytes {response!r}")

    with patch("roborock.devices.traits.v1.map_content.MapParser.parse", side_effect=parse_data):
        yield


async def test_discover_home_empty_cache(
    status_trait: StatusTrait,
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
    device_cache: DeviceCache,
    web_api_client: AsyncMock,
) -> None:
    """Test discovering home when cache is empty."""
    # Setup mocks for the discovery process
    mock_rpc_channel.send_command.side_effect = [
        UPDATED_STATUS_MAP_123,  # Status after switching to map 123
        ROOM_MAPPING_DATA_MAP_123,  # Rooms for map 123
        UPDATED_STATUS_MAP_0,  # Status after switching back to map 0
        ROOM_MAPPING_DATA_MAP_0,  # Rooms for map 0
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,  # Multi maps list
        {},  # LOAD_MULTI_MAP response for map 123
        {},  # LOAD_MULTI_MAP response back to map 0
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_2,  # Map bytes for 123
        MAP_BYTES_RESPONSE_1,  # Map bytes for 0
    ]
    # We have an empty home data so the room list gets loaded
    web_api_client.get_rooms.return_value = [
        HomeDataRoom(id=2362048, name="Example room 1"),
        HomeDataRoom(id=2362044, name="Example room 2"),
        HomeDataRoom(id=2362041, name="Example room 3"),
    ]

    # Before discovery, no cache should exist
    assert home_trait.home_map_info is None
    assert home_trait.current_map_data is None

    # Perform home discovery
    await home_trait.discover_home()

    # Verify cache is populated
    assert home_trait.home_map_info is not None
    assert len(home_trait.home_map_info) == 2

    # Check map 0 data
    map_0_data = home_trait.home_map_info[0]
    assert map_0_data.map_flag == 0
    assert map_0_data.name == "Ground Floor"
    assert len(map_0_data.rooms) == 2
    assert map_0_data.rooms[0].segment_id == 16
    assert map_0_data.rooms[0].iot_id == "2362048"
    assert map_0_data.rooms[0].name == "Example room 1"
    assert map_0_data.rooms[1].segment_id == 17
    assert map_0_data.rooms[1].iot_id == "2362044"
    assert map_0_data.rooms[1].name == "Example room 2"

    map_0_content = home_trait.home_map_content[0]
    assert map_0_content is not None
    assert map_0_content.image_content == TEST_IMAGE_CONTENT_1
    assert map_0_content.map_data is not None

    # Check map 123 data
    map_123_data = home_trait.home_map_info[123]
    assert map_123_data.map_flag == 123
    assert map_123_data.name == "Second Floor"
    assert len(map_123_data.rooms) == 2
    assert map_123_data.rooms[0].segment_id == 18
    assert map_123_data.rooms[0].iot_id == "2362041"
    assert map_123_data.rooms[0].name == "Example room 3"
    assert map_123_data.rooms[1].segment_id == 19
    assert map_123_data.rooms[1].iot_id == "2362042"
    assert map_123_data.rooms[1].name == "Room 19"  # Not in mock home data

    map_123_content = home_trait.home_map_content[123]
    assert map_123_content is not None
    assert map_123_content.image_content == TEST_IMAGE_CONTENT_2
    assert map_123_content.map_data is not None

    # Verify current map data is accessible
    current_map_data = home_trait.current_map_data
    assert current_map_data is not None
    assert current_map_data.map_flag == 0
    assert current_map_data.name == "Ground Floor"
    assert [room.name for room in home_trait.current_rooms] == [
        "Example room 1",
        "Example room 2",
    ]

    # Verify the persistent cache has been updated
    device_cache_data = await device_cache.get()
    assert device_cache_data.home_map_info is not None
    assert len(device_cache_data.home_map_info) == 2
    assert device_cache_data.home_map_content_base64 is not None
    assert len(device_cache_data.home_map_content_base64) == 2


@pytest.mark.parametrize(
    "device_cache_data",
    [
        DeviceCacheData(
            home_map_info={0: CombinedMapInfo(map_flag=0, name="Dummy", rooms=[])},
            home_map_content_base64={0: base64.b64encode(MAP_BYTES_RESPONSE_1).decode("utf-8")},
        ),
    ],
)
async def test_discover_home_with_existing_cache(
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    device_cache_data: DeviceCacheData,
    device_cache: DeviceCache,
) -> None:
    """Test that discovery is skipped when cache already exists."""
    # Pre-populate the cache
    await device_cache.set(device_cache_data)

    # Call discover_home
    await home_trait.discover_home()

    # Verify no RPC calls were made (discovery was skipped)
    assert mock_rpc_channel.send_command.call_count == 0
    assert mock_mqtt_rpc_channel.send_command.call_count == 0

    # Verify cache was loaded from storage
    assert home_trait.home_map_info == {0: CombinedMapInfo(map_flag=0, name="Dummy", rooms=[])}
    assert home_trait.home_map_content
    assert home_trait.home_map_content.keys() == {0}
    map_0_content = home_trait.home_map_content[0]
    assert map_0_content is not None
    assert map_0_content.image_content == TEST_IMAGE_CONTENT_1
    assert map_0_content.map_data is not None


async def test_existing_home_cache_invalid_bytes(
    home_trait: HomeTrait,
    device_cache: DeviceCache,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
) -> None:
    """Test that discovery is skipped when cache already exists."""
    # Pre-populate the cache.
    cache_data = await device_cache.get()
    cache_data.home_map_info = {0: CombinedMapInfo(map_flag=0, name="Dummy", rooms=[])}
    # We override the map bytes parser to raise an exception above.
    cache_data.home_map_content_base64 = {0: base64.b64encode(MAP_BYTES_RESPONSE_1).decode("utf-8")}
    await device_cache.set(cache_data)

    # Setup mocks for the discovery process
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,  # Rooms for the single map
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_SINGLE_MAP_DATA,  # Single map list
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,  # Map bytes for the single map
    ]

    # Call discover_home. First attempt raises an exception then loading from the server
    # produes a valid result.
    with patch(
        "roborock.devices.traits.v1.map_content.MapParser.parse",
        side_effect=[
            RoborockException("Invalid map bytes"),
            ParsedMapData(
                image_content=TEST_IMAGE_CONTENT_2,
                map_data=MagicMock(),
            ),
        ],
    ):
        await home_trait.discover_home()

    # Verify cache was loaded from storage. The map was re-fetched from storage.
    assert home_trait.home_map_info
    assert home_trait.home_map_info.keys() == {0}
    assert home_trait.home_map_info[0].name == "Only Floor"
    assert home_trait.home_map_content
    assert home_trait.home_map_content.keys() == {0}
    map_0_content = home_trait.home_map_content[0]
    assert map_0_content is not None
    assert map_0_content.image_content == TEST_IMAGE_CONTENT_2
    assert map_0_content.map_data is not None


async def test_discover_home_no_maps(
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
) -> None:
    """Test discovery when no maps are available."""
    # Setup mock to return empty maps list
    mock_mqtt_rpc_channel.send_command.side_effect = [
        # Discover home
        [{"max_multi_map": 0, "max_bak_map": 0, "multi_map_count": 0, "map_info": []}],
        # Refresh
        [{"max_multi_map": 0, "max_bak_map": 0, "multi_map_count": 0, "map_info": []}],
    ]

    # Discover home should not change anything
    await home_trait.discover_home()
    assert home_trait.current_map_data is None
    assert home_trait.home_map_info == {}
    assert home_trait.home_map_content == {}
    assert home_trait.current_rooms == []

    # Refresh should not change anything
    await home_trait.refresh()
    assert home_trait.current_map_data is None
    assert home_trait.home_map_info == {}
    assert home_trait.home_map_content == {}
    assert home_trait.current_rooms == []


async def test_refresh_updates_current_map_cache(
    device_cache: DeviceCache,
    status_trait: StatusTrait,
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
) -> None:
    """Test that refresh updates the cache for the current map."""
    # Pre-populate cache with some data
    cache_data = await device_cache.get()
    cache_data.home_map_info = {0: CombinedMapInfo(map_flag=0, name="Old Ground Floor", rooms=[])}
    cache_data.home_map_content_base64 = {
        0: base64.b64encode(MAP_BYTES_RESPONSE_2).decode()
    }  # Pre-existing different map bytes
    await device_cache.set(cache_data)
    await home_trait.discover_home()  # Load cache into trait
    # Verify initial cache state
    assert home_trait.home_map_info
    assert home_trait.home_map_info.keys() == {0}
    assert home_trait.home_map_info[0].name == "Old Ground Floor"
    assert len(home_trait.home_map_info[0].rooms) == 0
    assert home_trait.home_map_content
    assert home_trait.home_map_content.keys() == {0}
    assert home_trait.home_map_content[0].image_content == TEST_IMAGE_CONTENT_2

    # Setup mocks for refresh
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,  # Room mapping refresh
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,  # Maps refresh
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,  # Map bytes refresh
    ]

    # Perform refresh
    await home_trait.refresh()

    # Verify cache was updated for current map
    assert home_trait.home_map_info
    assert home_trait.home_map_info.keys() == {0}
    assert home_trait.home_map_info[0].name == "Ground Floor"
    assert len(home_trait.home_map_info[0].rooms) == 2
    # Verify map content
    assert home_trait.home_map_content
    assert home_trait.home_map_content.keys() == {0}
    assert home_trait.home_map_content[0].image_content == TEST_IMAGE_CONTENT_1


async def test_current_map_data_property(
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
) -> None:
    """Test current_map_data property returns correct data."""
    # Setup discovery
    mock_rpc_channel.send_command.side_effect = [
        UPDATED_STATUS_MAP_123,  # Status after switching to map 123
        ROOM_MAPPING_DATA_MAP_123,  # Rooms for map 123
        UPDATED_STATUS_MAP_0,  # Status after switching back to map 0
        ROOM_MAPPING_DATA_MAP_0,  # Rooms for map 0
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,  # Multi maps list
        {},  # LOAD_MULTI_MAP response for map 123
        {},  # LOAD_MULTI_MAP response back to map 0
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_2,  # Map bytes for 123
        MAP_BYTES_RESPONSE_1,  # Map bytes for 0
    ]

    await home_trait.discover_home()

    # Test current map data (should be map 0)
    current_data = home_trait.current_map_data
    assert current_data is not None
    assert current_data.map_flag == 0
    assert current_data.name == "Ground Floor"

    # Test when no cache exists
    home_trait._home_map_info = None
    assert home_trait.current_map_data is None


async def test_discover_home_device_busy_cleaning(
    status_trait: StatusTrait,
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
    device_cache: DeviceCache,
) -> None:
    """Test that discovery raises RoborockDeviceBusy when device is cleaning.

    This tests the initial failure scenario during discovery where the device
    is busy, then a retry attempt where the device is still busy, then finally
    a successful attempt to run discovery when the device is idle.
    """
    # Set the status trait state to cleaning
    status_trait.state = RoborockStateCode.cleaning

    # Attempt to discover home while cleaning
    with pytest.raises(RoborockDeviceBusy, match="Cannot perform home discovery while the device is cleaning"):
        await home_trait.discover_home()

    # Verify no RPC calls were made (discovery was prevented)
    assert mock_rpc_channel.send_command.call_count == 0
    assert mock_mqtt_rpc_channel.send_command.call_count == 0

    # Setup mocks for refresh
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,  # Room mapping refresh
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,  # Maps refresh
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,  # Map bytes refresh
    ]

    # Now attempt to refresh the device while cleaning. This should still fail
    # home discovery but allow refresh to update the current map.
    await home_trait.refresh()

    # Verify the home information is now populated
    current_data = home_trait.current_map_data
    assert current_data is not None
    assert current_data.map_flag == 0
    assert current_data.name == "Ground Floor"
    assert home_trait.home_map_info is not None
    assert home_trait.home_map_info.keys() == {0}
    assert home_trait.home_map_content is not None
    assert home_trait.home_map_content.keys() == {0}
    assert [room.name for room in home_trait.current_rooms] == [
        "Example room 1",
        "Example room 2",
    ]

    map_0_content = home_trait.home_map_content[0]
    assert map_0_content is not None
    assert map_0_content.image_content == TEST_IMAGE_CONTENT_1
    assert map_0_content.map_data is not None

    # Verify the persistent cache has not been updated since discovery
    # has not fully completed.
    cache_data = await device_cache.get()
    assert not cache_data.home_map_info
    assert not cache_data.home_map_content_base64

    # Set the status trait state to idle which will mean we can attempt discovery
    # on the next refresh. This should have the result of updating the
    # persistent cache which is verified below.
    status_trait.state = RoborockStateCode.idle

    # Setup mocks for the discovery process
    mock_rpc_channel.send_command.side_effect = [
        UPDATED_STATUS_MAP_123,  # Status after switching to map 123
        ROOM_MAPPING_DATA_MAP_123,  # Rooms for map 123
        UPDATED_STATUS_MAP_0,  # Status after switching back to map 0
        ROOM_MAPPING_DATA_MAP_0,  # Rooms for map 0
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,  # Multi maps list
        {},  # LOAD_MULTI_MAP response for map 123
        {},  # LOAD_MULTI_MAP response back to map 0
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_2,  # Map bytes for 123
        MAP_BYTES_RESPONSE_1,  # Map bytes for 0
    ]

    # Refreshing should now perform discovery successfully
    await home_trait.refresh()

    # Verify we now have all of the information populated from discovery
    assert home_trait.home_map_info is not None
    assert home_trait.home_map_info.keys() == {0, 123}
    assert home_trait.home_map_content is not None
    assert home_trait.home_map_content.keys() == {0, 123}
    assert [room.name for room in home_trait.current_rooms] == [
        "Room 16",
        "Room 17",
    ]

    # Verify the persistent cache has been updated
    device_cache_data = await device_cache.get()
    assert device_cache_data.home_map_info is not None
    assert len(device_cache_data.home_map_info) == 2
    assert device_cache_data.home_map_content_base64 is not None
    assert len(device_cache_data.home_map_content_base64) == 2


async def test_refresh_falls_back_when_map_switch_action_locked(
    status_trait: StatusTrait,
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
    device_cache: DeviceCache,
) -> None:
    """Test that refresh falls back to current map when map switching is locked."""
    # Discovery attempt: we can list maps, but switching maps fails with -10007.
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,  # Maps refresh during discover_home()
        RoborockInvalidStatus({"code": -10007, "message": "invalid status"}),  # LOAD_MULTI_MAP action locked
        MULTI_MAP_LIST_DATA,  # Maps refresh during refresh() fallback
    ]

    # Fallback refresh should still be able to refresh the current map.
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,  # Rooms for current map
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,  # Map bytes for current map
    ]

    await home_trait.refresh()

    current_data = home_trait.current_map_data
    assert current_data is not None
    assert current_data.map_flag == 0
    assert current_data.name == "Ground Floor"

    assert home_trait.home_map_info is not None
    assert home_trait.home_map_info.keys() == {0}
    assert home_trait.home_map_content is not None
    assert home_trait.home_map_content.keys() == {0}
    map_0_content = home_trait.home_map_content[0]
    assert map_0_content.image_content == TEST_IMAGE_CONTENT_1

    # Discovery did not complete, so the persistent cache should not be updated.
    cache_data = await device_cache.get()
    assert not cache_data.home_map_info
    assert not cache_data.home_map_content_base64


async def test_single_map_no_switching(
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
) -> None:
    """Test that single map discovery doesn't trigger map switching."""
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,  # Rooms for the single map
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_SINGLE_MAP_DATA,  # Single map list
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,  # Map bytes for the single map
    ]

    await home_trait.discover_home()

    # Verify cache is populated
    assert home_trait.home_map_info is not None
    assert home_trait.home_map_info.keys() == {0}
    assert home_trait.home_map_content is not None
    assert home_trait.home_map_content.keys() == {0}

    # Verify no LOAD_MULTI_MAP commands were sent (no map switching)
    load_map_calls = [
        call
        for call in mock_mqtt_rpc_channel.send_command.call_args_list
        if call[1].get("command") == RoborockCommand.LOAD_MULTI_MAP
    ]
    assert len(load_map_calls) == 0


async def test_refresh_map_info_prefers_map_info_names_and_adds_missing_rooms(
    home_trait: HomeTrait,
    rooms_trait: RoomsTrait,
) -> None:
    """Test room merge behavior in _refresh_map_info.

    This test verifies:
    1. Existing map_info room names are preserved.
    2. Missing rooms from RoomsTrait are added.
    3. map_info fallback names are replaced by RoomsTrait names.
    """
    map_info = MultiMapsListMapInfo(
        map_flag=0,
        name="Test Map",
        rooms=[
            MultiMapsListRoom(
                id=16,
                iot_name_id="2362048",
                iot_name="Kitchen from map_info",
            ),
            MultiMapsListRoom(
                id=19,
                iot_name_id="2362042",
                iot_name="Bedroom from map_info",
            ),
            MultiMapsListRoom(
                id=20,
                iot_name_id="9999001",
                iot_name=None,
            ),
        ],
    )

    # Mock rooms_trait to return multiple rooms covering all scenarios:
    # - segment_id 16 with fallback name: exists in map_info, should NOT override
    # - segment_id 19 with valid name: exists in map_info, should NOT override
    # - segment_id 17 with fallback name: not in map_info, should be added
    # - segment_id 18 with valid name: not in map_info, should be added
    # - segment_id 20 with valid name: overrides map_info fallback "Room 20"
    rooms_trait.rooms = [
        NamedRoomMapping(segment_id=16, iot_id="2362048"),  # Will override even with no name
        NamedRoomMapping(
            segment_id=19, iot_id="2362042", raw_name="Updated Bedroom Name"
        ),  # Exists in map_info, should not override
        NamedRoomMapping(segment_id=17, iot_id="2362044"),  # Not in map_info, should be added
        NamedRoomMapping(
            segment_id=18, iot_id="2362041", raw_name="Example room 3"
        ),  # Not in map_info, should be added
        NamedRoomMapping(segment_id=20, iot_id="9999001", raw_name="Office from rooms_trait"),
    ]

    # Mock rooms_trait.refresh to prevent actual device calls
    with patch.object(rooms_trait, "refresh", new_callable=AsyncMock):
        result = await home_trait._refresh_map_info(map_info)

    assert result.map_flag == 0
    assert result.name == "Test Map"
    assert len(result.rooms) == 5

    # Sort rooms by segment_id for consistent assertions
    sorted_rooms = sorted(result.rooms, key=lambda r: r.segment_id)

    # Room 16: from map_info is overridden by rooms_trait
    assert sorted_rooms[0].segment_id == 16
    assert sorted_rooms[0].name == "Room 16"
    assert sorted_rooms[0].iot_id == "2362048"

    # Room 17: from rooms_trait with fallback name, added because not in map_info
    assert sorted_rooms[1].segment_id == 17
    assert sorted_rooms[1].name == "Room 17"
    assert sorted_rooms[1].iot_id == "2362044"

    # Room 18: from rooms_trait with valid name, added because not in map_info
    assert sorted_rooms[2].segment_id == 18
    assert sorted_rooms[2].name == "Example room 3"
    assert sorted_rooms[2].iot_id == "2362041"

    # Room 19: from map_info, kept (not overridden by rooms_trait)
    assert sorted_rooms[3].segment_id == 19
    assert sorted_rooms[3].name == "Updated Bedroom Name"
    assert sorted_rooms[3].iot_id == "2362042"

    # Room 20: map_info fallback name replaced by rooms_trait name
    assert sorted_rooms[4].segment_id == 20
    assert sorted_rooms[4].name == "Office from rooms_trait"
    assert sorted_rooms[4].iot_id == "9999001"


async def test_home_trait_listener_notifications(
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
    device_cache: DeviceCache,
    web_api_client: AsyncMock,
) -> None:
    """Test that listener callbacks are called when home discovery/updates occur."""
    mock_callback = MagicMock()
    unsub = home_trait.add_update_listener(mock_callback)
    # Callback should NOT be executed immediately on subscription since home_map_info is None
    assert mock_callback.call_count == 0

    # 1. Test notification on empty cache discovery
    mock_rpc_channel.send_command.side_effect = [
        UPDATED_STATUS_MAP_123,
        ROOM_MAPPING_DATA_MAP_123,
        UPDATED_STATUS_MAP_0,
        ROOM_MAPPING_DATA_MAP_0,
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,
        {},
        {},
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_2,
        MAP_BYTES_RESPONSE_1,
    ]
    web_api_client.get_rooms.return_value = [
        HomeDataRoom(id=2362048, name="Example room 1"),
        HomeDataRoom(id=2362044, name="Example room 2"),
    ]

    await home_trait.discover_home()
    # Mock callback should have been called during discovery
    assert mock_callback.call_count > 0
    mock_callback.reset_mock()

    # 2. Test that registering a new listener now (when cache is populated) executes immediately
    mock_callback_immediate = MagicMock()
    unsub_immediate = home_trait.add_update_listener(mock_callback_immediate)
    assert mock_callback_immediate.call_count == 1
    unsub_immediate()

    # 3. Test notification on cached discovery
    # Re-run discover_home (which skips API calls and loads from cache)
    await home_trait.discover_home()
    assert mock_callback.call_count == 1
    mock_callback.reset_mock()

    # 4. Test notification on refresh
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,
    ]
    await home_trait.refresh()
    assert mock_callback.call_count > 0

    # Unsubscribe and verify it is no longer called
    mock_callback.reset_mock()
    unsub()
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,
    ]
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_DATA,
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,
    ]
    await home_trait.refresh()
    assert mock_callback.call_count == 0


async def test_home_trait_map_eviction(
    home_trait: HomeTrait,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
    device_cache: DeviceCache,
    web_api_client: AsyncMock,
) -> None:
    """Test that maps deleted from the device are evicted from cache during refresh."""
    # Pre-populate cache with maps 0 and 123
    device_cache_data = DeviceCacheData(
        home_map_info={
            0: CombinedMapInfo(map_flag=0, name="Ground Floor", rooms=[]),
            123: CombinedMapInfo(map_flag=123, name="Second Floor", rooms=[]),
        },
        home_map_content_base64={
            0: base64.b64encode(MAP_BYTES_RESPONSE_1).decode("utf-8"),
            123: base64.b64encode(MAP_BYTES_RESPONSE_2).decode("utf-8"),
        },
    )
    await device_cache.set(device_cache_data)
    await home_trait.discover_home()

    assert home_trait.home_map_info is not None
    assert home_trait.home_map_content is not None
    assert len(home_trait.home_map_info) == 2
    assert len(home_trait.home_map_content) == 2

    # Set up listener callback
    mock_callback = MagicMock()
    home_trait.add_update_listener(mock_callback)
    assert mock_callback.call_count == 1
    mock_callback.reset_mock()

    # Mock maps_trait.refresh so that only map 0 is returned (map 123 deleted)
    mock_mqtt_rpc_channel.send_command.side_effect = [
        MULTI_MAP_LIST_SINGLE_MAP_DATA,  # Only has map 0
    ]
    mock_rpc_channel.send_command.side_effect = [
        ROOM_MAPPING_DATA_MAP_0,
    ]
    mock_map_rpc_channel.send_command.side_effect = [
        MAP_BYTES_RESPONSE_1,
    ]

    await home_trait.refresh()

    # Verify map 123 is excluded from memory cache
    assert home_trait.home_map_info is not None
    assert home_trait.home_map_content is not None
    assert 123 not in home_trait.home_map_info
    assert 123 not in home_trait.home_map_content
    assert 0 in home_trait.home_map_info

    # Verify listener notified
    assert mock_callback.call_count > 0



"""Trait for managing maps and room mappings on Roborock devices.

New datatypes are introduced here to manage the additional information associated
with maps and rooms, such as map names and room names. These override the
base container datatypes to add additional fields.
"""

import logging

from roborock.data import MultiMapsList, MultiMapsListMapInfo
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand

from .status import StatusTrait

_LOGGER = logging.getLogger(__name__)


class MultiMapsListConverter(common.V1TraitDataConverter):
    """Converters responses to MultiMapsList."""

    def convert(self, response: common.V1ResponseData) -> MultiMapsList:
        """Parse the response from the device into a MapsTrait instance.

        This overrides the base implementation to handle the specific
        response format for the multi maps list. This is needed because we have
        a custom constructor that requires the StatusTrait.
        """
        if not isinstance(response, list):
            raise ValueError(f"Unexpected MapsTrait response format: {response!r}")
        response = response[0]
        if not isinstance(response, dict):
            raise ValueError(f"Unexpected MapsTrait response format: {response!r}")
        return MultiMapsList.from_dict(response)


@common.mqtt_rpc_channel
class MapsTrait(MultiMapsList, common.V1TraitMixin):
    """Trait for managing the maps of Roborock devices.

    A device may have multiple maps, each identified by a unique map_flag.
    Each map can have multiple rooms associated with it, in a `RoomMapping`.

    The MapsTrait depends on the StatusTrait to determine the currently active
    map. It is the responsibility of the caller to ensure that the StatusTrait
    is up to date before using this trait. However, there is a possibility of
    races if another client changes the current map between the time the
    StatusTrait is refreshed and when the MapsTrait is used. This is mitigated
    by the fact that the map list is unlikely to change frequently, and the
    current map is only changed when the user explicitly switches maps.
    """

    command = RoborockCommand.GET_MULTI_MAPS_LIST
    converter = MultiMapsListConverter()

    def __init__(self, status_trait: StatusTrait) -> None:
        """Initialize the MapsTrait.

        We keep track of the StatusTrait to ensure we have the latest
        status information when dealing with maps.
        """
        super().__init__()
        self._status_trait = status_trait

    @property
    def current_map(self) -> int | None:
        """Returns the currently active map (map_flag), if available."""
        return self._status_trait.current_map

    @property
    def current_map_info(self) -> MultiMapsListMapInfo | None:
        """Returns the currently active map info, if available."""
        if (current_map := self.current_map) is None or self.map_info is None:
            return None
        for map_info in self.map_info:
            if map_info.map_flag == current_map:
                return map_info
        return None

    async def set_current_map(self, map_flag: int) -> None:
        """Update the current map of the device by it's map_flag id."""
        await self.rpc_channel.send_command(RoborockCommand.LOAD_MULTI_MAP, params=[map_flag])
        # Refresh our status to make sure it reflects the new map
        await self._status_trait.refresh()



"""Tests for the health manager."""

import datetime
from unittest.mock import AsyncMock, patch

from roborock.mqtt.health_manager import HealthManager


async def test_health_manager_restart_called_after_timeouts() -> None:
    """Test that the health manager calls restart after consecutive timeouts."""
    restart = AsyncMock()
    health_manager = HealthManager(restart=restart)

    await health_manager.on_timeout()
    await health_manager.on_timeout()
    restart.assert_not_called()

    await health_manager.on_timeout()
    restart.assert_called_once()


async def test_health_manager_success_resets_counter() -> None:
    """Test that a successful message resets the timeout counter."""
    restart = AsyncMock()
    health_manager = HealthManager(restart=restart)

    await health_manager.on_timeout()
    await health_manager.on_timeout()
    restart.assert_not_called()

    await health_manager.on_success()

    await health_manager.on_timeout()
    await health_manager.on_timeout()
    restart.assert_not_called()

    await health_manager.on_timeout()
    restart.assert_called_once()


async def test_cooldown() -> None:
    """Test that the health manager respects the restart cooldown."""
    restart = AsyncMock()
    health_manager = HealthManager(restart=restart)

    with patch("roborock.mqtt.health_manager.datetime") as mock_datetime:
        now = datetime.datetime(2023, 1, 1, 12, 0, 0)
        mock_datetime.datetime.now.return_value = now

        # Trigger first restart
        await health_manager.on_timeout()
        await health_manager.on_timeout()
        await health_manager.on_timeout()
        restart.assert_called_once()
        restart.reset_mock()

        # Advance time but stay within cooldown (30 mins)
        mock_datetime.datetime.now.return_value = now + datetime.timedelta(minutes=10)

        # Trigger timeouts again
        await health_manager.on_timeout()
        await health_manager.on_timeout()
        await health_manager.on_timeout()
        restart.assert_not_called()

        # Advance time past cooldown
        mock_datetime.datetime.now.return_value = now + datetime.timedelta(minutes=31)

        # Trigger timeouts again
        await health_manager.on_timeout()
        await health_manager.on_timeout()
        await health_manager.on_timeout()
        restart.assert_called_once()



"""Test cases for the containers module."""

import dataclasses
from dataclasses import dataclass
from typing import Any

import pytest

from roborock.data.containers import (
    HomeData,
    HomeDataDevice,
    RoborockBase,
    RoborockCategory,
    UserData,
    _camelize,
    _decamelize,
)
from roborock.roborock_message import RoborockDataProtocol, RoborockMessageProtocol
from tests.mock_data import (
    HOME_DATA_RAW,
    K_VALUE,
    LOCAL_KEY,
    USER_DATA,
)


@dataclass
class SimpleObject(RoborockBase):
    """Simple object for testing serialization."""

    name: str | None = None
    value: int | None = None


@dataclass
class ComplexObject(RoborockBase):
    """Complex object for testing serialization."""

    simple: SimpleObject | None = None
    items: list[str] | None = None
    value: int | None = None
    nested_dict: dict[str, SimpleObject] | None = None
    nested_list: list[SimpleObject] | None = None
    any: Any | None = None
    nested_int_dict: dict[int, SimpleObject] | None = None


@dataclass
class BoolFeatures(RoborockBase):
    """Complex object for testing serialization."""

    my_flag_supported: bool | None = None
    my_flag_2_supported: bool | None = None
    is_ces_2022_supported: bool | None = None


def test_simple_object() -> None:
    """Test serialization and deserialization of a simple object."""

    obj = SimpleObject(name="Test", value=42)
    serialized = obj.as_dict()
    assert serialized == {"name": "Test", "value": 42}
    deserialized = SimpleObject.from_dict(serialized)
    assert deserialized.name == "Test"
    assert deserialized.value == 42


def test_complex_object() -> None:
    """Test serialization and deserialization of a complex object."""
    simple = SimpleObject(name="Nested", value=100)
    obj = ComplexObject(
        simple=simple,
        items=["item1", "item2"],
        value=200,
        nested_dict={
            "nested1": SimpleObject(name="Nested1", value=1),
            "nested2": SimpleObject(name="Nested2", value=2),
        },
        nested_int_dict={
            10: SimpleObject(name="IntKey1", value=10),
        },
        nested_list=[SimpleObject(name="Nested3", value=3), SimpleObject(name="Nested4", value=4)],
        any="This can be anything",
    )
    serialized = obj.as_dict()
    assert serialized == {
        "simple": {"name": "Nested", "value": 100},
        "items": ["item1", "item2"],
        "value": 200,
        "nestedDict": {
            "nested1": {"name": "Nested1", "value": 1},
            "nested2": {"name": "Nested2", "value": 2},
        },
        "nestedIntDict": {
            10: {"name": "IntKey1", "value": 10},
        },
        "nestedList": [
            {"name": "Nested3", "value": 3},
            {"name": "Nested4", "value": 4},
        ],
        "any": "This can be anything",
    }
    deserialized = ComplexObject.from_dict(serialized)
    assert deserialized.simple.name == "Nested"
    assert deserialized.simple.value == 100
    assert deserialized.items == ["item1", "item2"]
    assert deserialized.value == 200
    assert deserialized.nested_dict == {
        "nested1": SimpleObject(name="Nested1", value=1),
        "nested2": SimpleObject(name="Nested2", value=2),
    }
    assert deserialized.nested_int_dict == {
        10: SimpleObject(name="IntKey1", value=10),
    }
    assert deserialized.nested_list == [
        SimpleObject(name="Nested3", value=3),
        SimpleObject(name="Nested4", value=4),
    ]
    assert deserialized.any == "This can be anything"


@pytest.mark.parametrize(
    ("data"),
    [
        {
            "nested_int_dict": {10: {"name": "IntKey1", "value": 10}},
        },
        {
            "nested_int_dict": {"10": {"name": "IntKey1", "value": 10}},
        },
    ],
)
def test_from_dict_key_types(data: dict) -> None:
    """Test serialization and deserialization of a complex object."""
    obj = ComplexObject.from_dict(data)
    assert obj.nested_int_dict == {
        10: SimpleObject(name="IntKey1", value=10),
    }


def test_ignore_unknown_keys() -> None:
    """Test that we don't fail on unknown keys."""
    data = {
        "ignored_key": "This key should be ignored",
        "name": "named_object",
        "value": 42,
    }
    deserialized = SimpleObject.from_dict(data)
    assert deserialized.name == "named_object"
    assert deserialized.value == 42


def test_user_data():
    ud = UserData.from_dict(USER_DATA)
    assert ud.uid == 123456
    assert ud.tokentype == "token_type"
    assert ud.token == "abc123"
    assert ud.rruid == "abc123"
    assert ud.region == "us"
    assert ud.country == "US"
    assert ud.countrycode == "1"
    assert ud.nickname == "user_nickname"
    assert ud.rriot.u == "user123"
    assert ud.rriot.s == "pass123"
    assert ud.rriot.h == "unknown123"
    assert ud.rriot.k == K_VALUE
    assert ud.rriot.r.r == "US"
    assert ud.rriot.r.a == "https://api-us.roborock.com"
    assert ud.rriot.r.m == "tcp://mqtt-us.roborock.com:8883"
    assert ud.rriot.r.l == "https://wood-us.roborock.com"
    assert ud.tuya_device_state == 2
    assert ud.avatarurl == "https://files.roborock.com/iottest/default_avatar.png"


def test_home_data():
    hd = HomeData.from_dict(HOME_DATA_RAW)
    assert hd.id == 123456
    assert hd.name == "My Home"
    assert hd.lon is None
    assert hd.lat is None
    assert hd.geo_name is None
    product = hd.products[0]
    assert product.id == "product-id-s7-maxv"
    assert product.name == "Roborock S7 MaxV"
    assert product.code == "a27"
    assert product.model == "roborock.vacuum.a27"
    assert product.icon_url is None
    assert product.attribute is None
    assert product.capability == 0
    assert product.category == RoborockCategory.VACUUM
    schema = product.schema
    assert schema[0].id == "101"
    assert schema[0].name == "rpc_request"
    assert schema[0].code == "rpc_request_code"
    assert schema[0].mode == "rw"
    assert schema[0].type == "RAW"
    assert schema[0].product_property is None
    assert schema[0].desc is None
    assert product.supported_schema_codes == {
        "additional_props",
        "battery",
        "charge_status",
        "drying_status",
        "error_code",
        "fan_power",
        "filter_life",
        "main_brush_life",
        "rpc_request_code",
        "rpc_response",
        "side_brush_life",
        "state",
        "task_cancel_in_motion",
        "task_cancel_low_power",
        "task_complete",
        "water_box_mode",
    }
    assert product.supported_schema_ids == {
        int(v)
        for v in (
            RoborockMessageProtocol.RPC_REQUEST,
            RoborockMessageProtocol.RPC_RESPONSE,
            RoborockDataProtocol.ERROR_CODE,
            RoborockDataProtocol.STATE,
            RoborockDataProtocol.BATTERY,
            RoborockDataProtocol.FAN_POWER,
            RoborockDataProtocol.WATER_BOX_MODE,
            RoborockDataProtocol.MAIN_BRUSH_WORK_TIME,
            RoborockDataProtocol.SIDE_BRUSH_WORK_TIME,
            RoborockDataProtocol.FILTER_WORK_TIME,
            RoborockDataProtocol.ADDITIONAL_PROPS,
            RoborockDataProtocol.TASK_COMPLETE,
            RoborockDataProtocol.TASK_CANCEL_LOW_POWER,
            RoborockDataProtocol.TASK_CANCEL_IN_MOTION,
            RoborockDataProtocol.CHARGE_STATUS,
            RoborockDataProtocol.DRYING_STATUS,
        )
    }

    device = hd.devices[0]
    assert device.duid == "abc123"
    assert device.name == "Roborock S7 MaxV"
    assert device.attribute is None
    assert device.active_time == 1672364449
    assert device.local_key == LOCAL_KEY
    assert device.runtime_env is None
    assert device.time_zone_id == "America/Los_Angeles"
    assert device.icon_url == "no_url"
    assert device.product_id == "product-id-s7-maxv"
    assert device.lon is None
    assert device.lat is None
    assert not device.share
    assert device.share_time is None
    assert device.online
    assert device.fv == "02.56.02"
    assert device.pv == "1.0"
    assert device.room_id == 2362003
    assert device.tuya_uuid is None
    assert not device.tuya_migrated
    assert device.extra == '{"RRPhotoPrivacyVersion": "1"}'
    assert device.sn == "abc123"
    assert device.feature_set == "2234201184108543"
    assert device.new_feature_set == "0000000000002041"
    # status = device.device_status
    # assert status.name ==
    assert device.silent_ota_switch
    assert hd.rooms[0].id == 2362048
    assert hd.rooms[0].name == "Example room 1"


def test_serialize_and_unserialize():
    ud = UserData.from_dict(USER_DATA)
    ud_dict = ud.as_dict()
    assert ud_dict == USER_DATA


def test_boolean_features() -> None:
    """Test serialization and deserialization of BoolFeatures."""
    obj = BoolFeatures(my_flag_supported=True, my_flag_2_supported=False, is_ces_2022_supported=True)
    serialized = obj.as_dict()
    assert serialized == {
        "myFlagSupported": True,
        "myFlag2Supported": False,
        "isCes2022Supported": True,
    }
    deserialized = BoolFeatures.from_dict(serialized)
    assert dataclasses.asdict(deserialized) == {
        "my_flag_supported": True,
        "my_flag_2_supported": False,
        "is_ces_2022_supported": True,
    }


@pytest.mark.parametrize(
    "input_str,expected",
    [
        ("simpleTest", "simple_test"),
        ("testValue", "test_value"),
        ("anotherExampleHere", "another_example_here"),
        ("isCes2022Supported", "is_ces_2022_supported"),
        ("isThreeDMappingInnerTestSupported", "is_three_d_mapping_inner_test_supported"),
    ],
)
def test_decamelize_function(input_str: str, expected: str) -> None:
    """Test the _decamelize function."""

    assert _decamelize(input_str) == expected
    assert _camelize(expected) == input_str


def test_offline_device() -> None:
    """Test that a HomeDataDevice response from an offline device is handled correctly."""
    data = {
        "duid": "xxxxxx",
        "name": "S6 Pure",
        "localKey": "yyyyy",
        "productId": "zzzzz",
        "activeTime": 1765277892,
        "timeZoneId": "Europe/Moscow",
        "iconUrl": "",
        "share": False,
        "online": False,
        "pv": "1.0",
        "tuyaMigrated": False,
        "extra": "{}",
        "deviceStatus": {},
        "silentOtaSwitch": False,
        "f": False,
    }
    device = HomeDataDevice.from_dict(data)
    assert device.duid == "xxxxxx"
    assert device.name == "S6 Pure"
    assert device.local_key == "yyyyy"
    assert device.product_id == "zzzzz"
    assert device.active_time == 1765277892
    assert device.time_zone_id == "Europe/Moscow"
    assert not device.online
    assert device.fv is None



"""Tests for the RoomMapping related functionality."""

from copy import deepcopy
from typing import Any
from unittest.mock import AsyncMock

import pytest

from roborock.data.containers import HomeData, HomeDataRoom, NamedRoomMapping
from roborock.devices.device import RoborockDevice
from roborock.devices.traits.v1.rooms import RoomsTrait
from roborock.devices.traits.v1.status import StatusTrait
from roborock.roborock_typing import RoborockCommand
from tests.devices.traits.v1.fixtures import HOME_DATA


@pytest.fixture
def status_trait(device: RoborockDevice) -> StatusTrait:
    """Create a StatusTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.status


@pytest.fixture
def rooms_trait(device: RoborockDevice) -> RoomsTrait:
    """Create a RoomsTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.rooms


# Rooms from mock_data.HOME_DATA
# {"id": 2362048, "name": "Example room 1"},
# {"id": 2362044, "name": "Example room 2"},
# {"id": 2362041, "name": "Example room 3"},
@pytest.mark.parametrize(
    ("room_mapping_data"),
    [
        ([[16, "2362048"], [17, "2362044"], [18, "2362041"]]),
        ([[16, "2362048", 6], [17, "2362044", 14], [18, "2362041", 13]]),
    ],
)
async def test_refresh_rooms_trait(
    rooms_trait: RoomsTrait,
    mock_rpc_channel: AsyncMock,
    room_mapping_data: list[Any],
) -> None:
    """Test successfully getting room mapping."""
    # Setup mock to return the sample room mapping
    mock_rpc_channel.send_command.side_effect = [room_mapping_data]
    # Before refresh, rooms should be empty
    assert not rooms_trait.rooms

    # Load the room mapping information
    await rooms_trait.refresh()

    # Verify the room mappings are now populated
    assert rooms_trait.rooms
    rooms = rooms_trait.rooms
    assert len(rooms) == 3

    assert rooms[0].segment_id == 16
    assert rooms[0].name == "Example room 1"
    assert rooms[0].iot_id == "2362048"

    assert rooms[1].segment_id == 17
    assert rooms[1].name == "Example room 2"
    assert rooms[1].iot_id == "2362044"

    assert rooms[2].segment_id == 18
    assert rooms[2].name == "Example room 3"
    assert rooms[2].iot_id == "2362041"

    # Verify the RPC call was made correctly
    assert mock_rpc_channel.send_command.call_count == 1
    mock_rpc_channel.send_command.assert_any_call(RoborockCommand.GET_ROOM_MAPPING)


async def test_refresh_unknown_room_names_overwrites_home_data(
    rooms_trait: RoomsTrait,
    web_api_client: AsyncMock,
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test web rooms are used to refresh home data for missing iot ids."""
    web_api_client.get_rooms.return_value = [
        HomeDataRoom(id=2362048, name="Living Room"),
        HomeDataRoom(id=2362044, name="Example room 2"),
        HomeDataRoom(id=2362041, name="Example room 3"),
        HomeDataRoom(id=9999999, name="Office"),
    ]

    room_mapping_data = [[16, "2362048"], [17, "9999999"]]
    mock_rpc_channel.send_command.side_effect = [room_mapping_data]

    await rooms_trait.refresh()

    assert rooms_trait.rooms
    assert rooms_trait.rooms[0] == NamedRoomMapping(segment_id=16, iot_id="2362048", raw_name="Living Room")
    assert rooms_trait.rooms[1] == NamedRoomMapping(segment_id=17, iot_id="9999999", raw_name="Office")

    home_data_rooms = {str(room.id): room.name for room in rooms_trait._home_data.rooms or ()}
    assert home_data_rooms["2362048"] == "Living Room"
    assert home_data_rooms["9999999"] == "Office"


async def test_refresh_unknown_room_names_web_api_called_once(
    rooms_trait: RoomsTrait,
    web_api_client: AsyncMock,
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test unknown room IDs trigger one web lookup per iot_id."""
    web_api_client.get_rooms.return_value = [
        HomeDataRoom(id=9999911, name="Living Room"),
    ]

    room_mapping_data = [[16, "9999911"]]
    mock_rpc_channel.send_command.side_effect = [room_mapping_data, room_mapping_data]

    await rooms_trait.refresh()
    assert rooms_trait.rooms
    assert rooms_trait.rooms[0].name == "Living Room"

    await rooms_trait.refresh()
    assert rooms_trait.rooms
    assert rooms_trait.rooms[0].name == "Living Room"
    web_api_client.get_rooms.assert_called_once()


async def test_refresh_unknown_room_names_unresolved_uses_room_fallback(
    rooms_trait: RoomsTrait,
    web_api_client: AsyncMock,
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test unresolved unknown names use Room fallback in RoomsTrait."""
    web_api_client.get_rooms.return_value = []
    room_mapping_data = [[33, "9999922"]]
    mock_rpc_channel.send_command.side_effect = [room_mapping_data, room_mapping_data]

    await rooms_trait.refresh()
    assert rooms_trait.rooms
    assert rooms_trait.rooms[0] == NamedRoomMapping(segment_id=33, iot_id="9999922")
    assert rooms_trait.rooms[0].name == "Room 33"

    await rooms_trait.refresh()
    assert rooms_trait.rooms
    assert rooms_trait.rooms[0] == NamedRoomMapping(segment_id=33, iot_id="9999922")
    assert rooms_trait.rooms[0].name == "Room 33"
    web_api_client.get_rooms.assert_called_once()


async def test_refresh_unknown_room_names_called_again_for_new_unknown_room(
    rooms_trait: RoomsTrait,
    web_api_client: AsyncMock,
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test get_rooms is called again when a new unknown room appears."""
    room_mapping_data_1 = [[16, "9999601"]]
    room_mapping_data_2 = [[16, "9999601"], [17, "9999602"]]
    mock_rpc_channel.send_command.side_effect = [room_mapping_data_1, room_mapping_data_2]
    web_api_client.get_rooms.return_value = []

    await rooms_trait.refresh()
    assert rooms_trait.rooms
    assert sorted(room.name for room in rooms_trait.rooms) == ["Room 16"]

    await rooms_trait.refresh()
    assert rooms_trait.rooms
    assert sorted(room.name for room in rooms_trait.rooms) == ["Room 16", "Room 17"]
    assert web_api_client.get_rooms.call_count == 2


async def test_refresh_unknown_room_names_called_again_for_new_unknown_iot_id_same_segment(
    rooms_trait: RoomsTrait,
    web_api_client: AsyncMock,
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test get_rooms is called again for a new unknown iot_id on the same segment."""
    room_mapping_data_1 = [[16, "9999501"]]
    room_mapping_data_2 = [[16, "9999502"]]
    mock_rpc_channel.send_command.side_effect = [room_mapping_data_1, room_mapping_data_2]
    web_api_client.get_rooms.return_value = []

    await rooms_trait.refresh()
    await rooms_trait.refresh()

    assert web_api_client.get_rooms.call_count == 2


async def test_refresh_unknown_room_names_failure_falls_back_to_room_segment_id(
    rooms_trait: RoomsTrait,
    web_api_client: AsyncMock,
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test get_rooms failure gracefully falls back to Room {segment_id}."""
    room_mapping_data = [[16, "9999401"]]
    mock_rpc_channel.send_command.side_effect = [room_mapping_data]
    web_api_client.get_rooms.side_effect = Exception("API error")

    await rooms_trait.refresh()

    assert rooms_trait.rooms
    assert rooms_trait.rooms[0] == NamedRoomMapping(segment_id=16, iot_id="9999401")
    assert rooms_trait.rooms[0].name == "Room 16"
    web_api_client.get_rooms.assert_called_once()


def _build_shared_home_data() -> HomeData:
    home_data = deepcopy(HOME_DATA)
    home_data.received_devices = [home_data.devices.pop(0)]
    return home_data


@pytest.mark.parametrize("trait_home_data", [_build_shared_home_data()], indirect=True)
async def test_refresh_shared_room_names_use_shared_device_rooms(
    rooms_trait: RoomsTrait,
    trait_home_data: HomeData,
    web_api_client: AsyncMock,
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test shared devices resolve room names via the shared-device room list."""
    assert trait_home_data.received_devices
    assert not rooms_trait.rooms

    web_api_client.get_shared_device_rooms.return_value = [
        HomeDataRoom(id=9999999, name="Office"),
    ]
    room_mapping_data = [[16, "2362048"], [17, "9999999"]]
    mock_rpc_channel.send_command.side_effect = [room_mapping_data]

    await rooms_trait.refresh()

    assert rooms_trait.rooms == [
        NamedRoomMapping(segment_id=16, iot_id="2362048", raw_name="Example room 1"),
        NamedRoomMapping(segment_id=17, iot_id="9999999", raw_name="Office"),
    ]
    web_api_client.get_shared_device_rooms.assert_called_once_with(rooms_trait._device_uid)
    web_api_client.get_rooms.assert_not_called()



"""Tests for the StatusTrait class."""

import asyncio
from typing import cast
from unittest.mock import AsyncMock

import pytest

from roborock import (
    CleaningMode,
    CleanRoutes,
    VacuumModes,
    WaterModes,
    get_cleaning_mode_parameters,
    get_current_cleaning_mode,
    resolve_cleaning_mode,
)
from roborock.data import SHORT_MODEL_TO_ENUM
from roborock.data.v1 import (
    RoborockStateCode,
)
from roborock.device_features import DeviceFeatures
from roborock.devices.device import RoborockDevice
from roborock.devices.traits.v1.device_features import DeviceFeaturesTrait
from roborock.devices.traits.v1.status import StatusTrait
from roborock.exceptions import RoborockException, RoborockParsingException, RoborockUnsupportedFeature
from roborock.roborock_message import RoborockDataProtocol
from roborock.roborock_typing import RoborockCommand
from tests import mock_data
from tests.mock_data import STATUS


@pytest.fixture
def status_trait(device: RoborockDevice) -> StatusTrait:
    """Create a StatusTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.status


def _create_cleaning_mode_status_trait(**feature_overrides: bool) -> StatusTrait:
    """Create a status trait with mop-capable V1 features for cleaning mode tests."""
    short_model = mock_data.A27_PRODUCT_DATA["model"].split(".")[-1]
    features = DeviceFeatures.from_feature_flags(
        new_feature_info=0,
        new_feature_info_str="",
        feature_info=[],
        product_nickname=SHORT_MODEL_TO_ENUM[short_model],
    )
    features.is_support_water_mode = True
    features.is_pure_clean_mop_supported = True
    features.is_customized_clean_supported = True
    features.is_clean_route_setting_supported = True
    for feature_name, value in feature_overrides.items():
        if not hasattr(features, feature_name):
            raise AttributeError(f"Unknown DeviceFeatures override: {feature_name}")
        setattr(features, feature_name, value)
    return StatusTrait(cast(DeviceFeaturesTrait, features), region="us")


async def test_refresh_status(status_trait: StatusTrait, mock_rpc_channel: AsyncMock) -> None:
    """Test successfully refreshing status."""
    mock_rpc_channel.send_command.return_value = [STATUS]

    await status_trait.refresh()

    assert status_trait.battery == 100
    assert status_trait.state == RoborockStateCode.charging
    assert status_trait.fan_power == 102
    assert status_trait.fan_speed_name == "balanced"
    assert status_trait.fan_speed_name in status_trait.fan_speed_options
    mock_rpc_channel.send_command.assert_called_once_with(RoborockCommand.GET_STATUS)


async def test_refresh_status_dict_response(status_trait: StatusTrait, mock_rpc_channel: AsyncMock) -> None:
    """Test refreshing status when response is a dict instead of list."""
    mock_rpc_channel.send_command.return_value = STATUS

    await status_trait.refresh()

    assert status_trait.battery == 100
    assert status_trait.state == RoborockStateCode.charging
    mock_rpc_channel.send_command.assert_called_once_with(RoborockCommand.GET_STATUS)


async def test_refresh_status_propagates_exception(status_trait: StatusTrait, mock_rpc_channel: AsyncMock) -> None:
    """Test that exceptions from RPC channel are propagated."""
    mock_rpc_channel.send_command.side_effect = RoborockException("Communication error")

    with pytest.raises(RoborockException, match="Communication error"):
        await status_trait.refresh()


async def test_refresh_status_invalid_format(status_trait: StatusTrait, mock_rpc_channel: AsyncMock) -> None:
    """Test that invalid response format raises RoborockParsingException."""
    mock_rpc_channel.send_command.return_value = "invalid"

    with pytest.raises(RoborockParsingException, match="Unexpected StatusV2 response format"):
        await status_trait.refresh()


def test_none_values(status_trait: StatusTrait) -> None:
    """Test that none values are returned correctly."""
    status_trait.fan_power = None
    status_trait.water_box_mode = None
    status_trait.mop_mode = None
    assert status_trait.fan_speed_name is None
    assert status_trait.water_mode_name is None
    assert status_trait.mop_route_name is None


def test_options(status_trait: StatusTrait) -> None:
    """Test that fan_speed_options returns a list of options."""
    assert isinstance(status_trait.fan_speed_options, list)
    assert len(status_trait.fan_speed_options) > 0
    assert isinstance(status_trait.water_mode_options, list)
    assert len(status_trait.water_mode_options) > 0
    assert isinstance(status_trait.mop_route_options, list)
    assert len(status_trait.mop_route_options) > 0


def test_cleaning_mode_options() -> None:
    """Test the high-level cleaning mode options for the device."""
    status_trait = _create_cleaning_mode_status_trait()
    assert status_trait.cleaning_mode_options == [
        CleaningMode.VACUUM,
        CleaningMode.VAC_AND_MOP,
        CleaningMode.MOP,
        CleaningMode.CUSTOM,
    ]


@pytest.mark.parametrize(
    ("fan_power", "water_box_mode", "mop_mode", "expected_mode"),
    [
        (
            VacuumModes.BALANCED.code,
            WaterModes.STANDARD.code,
            CleanRoutes.STANDARD.code,
            CleaningMode.VAC_AND_MOP,
        ),
        (
            VacuumModes.BALANCED.code,
            WaterModes.OFF.code,
            CleanRoutes.STANDARD.code,
            CleaningMode.VACUUM,
        ),
        (
            VacuumModes.OFF.code,
            WaterModes.STANDARD.code,
            CleanRoutes.STANDARD.code,
            CleaningMode.MOP,
        ),
        (
            VacuumModes.CUSTOMIZED.code,
            WaterModes.STANDARD.code,
            CleanRoutes.STANDARD.code,
            CleaningMode.CUSTOM,
        ),
        (
            VacuumModes.BALANCED.code,
            WaterModes.SMART_MODE.code,
            CleanRoutes.STANDARD.code,
            CleaningMode.SMART_MODE,
        ),
    ],
)
def test_current_cleaning_mode(
    fan_power: int,
    water_box_mode: int,
    mop_mode: int,
    expected_mode: CleaningMode,
) -> None:
    """Test the current high-level cleaning mode classification."""
    status_trait = _create_cleaning_mode_status_trait(is_smart_clean_mode_set_supported=True)
    status_trait.fan_power = fan_power
    status_trait.water_box_mode = water_box_mode
    status_trait.mop_mode = mop_mode

    assert status_trait.current_cleaning_mode == expected_mode
    assert status_trait.current_cleaning_mode_name == expected_mode.value


def test_current_cleaning_mode_with_brush_up_mop() -> None:
    """Test brush-up mop-only classification on supported devices."""
    status_trait = _create_cleaning_mode_status_trait(is_support_main_brush_up_down_supported=True)
    status_trait.fan_power = VacuumModes.OFF_RAISE_MAIN_BRUSH.code
    status_trait.water_box_mode = WaterModes.STANDARD.code
    status_trait.mop_mode = CleanRoutes.STANDARD.code

    assert status_trait.current_cleaning_mode == CleaningMode.MOP


def test_current_cleaning_mode_accepts_enums() -> None:
    """Test direct enum inputs are resolved before classification."""
    status_trait = _create_cleaning_mode_status_trait(is_smart_clean_mode_set_supported=True)

    assert (
        get_current_cleaning_mode(
            clean_mode=VacuumModes.BALANCED,
            water_mode=WaterModes.SMART_MODE,
            mop_mode=CleanRoutes.STANDARD,
            features=status_trait._device_features_trait,
        )
        == CleaningMode.SMART_MODE
    )


def test_current_cleaning_mode_none() -> None:
    """Test that incomplete status values do not classify a cleaning mode."""
    status_trait = _create_cleaning_mode_status_trait()
    status_trait.fan_power = None
    assert status_trait.current_cleaning_mode is None
    assert status_trait.current_cleaning_mode_name is None


def test_current_cleaning_mode_without_mop_route_status() -> None:
    """Test older V1 devices can classify cleaning mode without mop route status."""
    status_trait = _create_cleaning_mode_status_trait(
        is_clean_route_setting_supported=False,
        is_customized_clean_supported=False,
    )
    status_trait.fan_power = VacuumModes.BALANCED.code
    status_trait.water_box_mode = WaterModes.OFF.code
    status_trait.mop_mode = None

    assert status_trait.current_cleaning_mode == CleaningMode.VACUUM


def test_get_cleaning_mode_parameters() -> None:
    """Test payload generation for supported high-level cleaning modes."""
    status_trait = _create_cleaning_mode_status_trait()
    assert get_cleaning_mode_parameters(CleaningMode.VACUUM, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.BALANCED.code,
            "water_box_mode": WaterModes.OFF.code,
            "mop_mode": CleanRoutes.STANDARD.code,
        }
    ]
    assert get_cleaning_mode_parameters(resolve_cleaning_mode("custom"), status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.CUSTOMIZED.code,
            "water_box_mode": WaterModes.CUSTOMIZED.code,
            "mop_mode": CleanRoutes.CUSTOMIZED.code,
        }
    ]


def test_get_cleaning_mode_parameters_unsupported() -> None:
    """Test unsupported cleaning modes raise a clear error."""
    status_trait = _create_cleaning_mode_status_trait()
    with pytest.raises(RoborockUnsupportedFeature, match="not supported"):
        get_cleaning_mode_parameters(CleaningMode.SMART_MODE, status_trait._device_features_trait)


def test_get_cleaning_mode_parameters_invalid_name() -> None:
    """Test invalid cleaning mode names raise RoborockUnsupportedFeature."""
    with pytest.raises(RoborockUnsupportedFeature, match="not supported"):
        resolve_cleaning_mode("invalid_mode")


async def test_set_cleaning_mode(
    mock_rpc_channel: AsyncMock,
) -> None:
    """Test setting the high-level cleaning mode."""
    status_trait = _create_cleaning_mode_status_trait()
    status_trait._rpc_channel = mock_rpc_channel  # type: ignore[assignment]
    await status_trait.set_cleaning_mode(CleaningMode.CUSTOM)

    mock_rpc_channel.send_command.assert_called_once_with(
        RoborockCommand.SET_CLEAN_MOTOR_MODE,
        params=[
            {
                "fan_power": VacuumModes.CUSTOMIZED.code,
                "water_box_mode": WaterModes.CUSTOMIZED.code,
                "mop_mode": CleanRoutes.CUSTOMIZED.code,
            }
        ],
    )


def test_cleaning_mode_options_with_smart_mode() -> None:
    """Test SmartPlan support is reflected in the available options."""
    status_trait = _create_cleaning_mode_status_trait(is_smart_clean_mode_set_supported=True)

    assert status_trait.cleaning_mode_options == [
        CleaningMode.VACUUM,
        CleaningMode.VAC_AND_MOP,
        CleaningMode.MOP,
        CleaningMode.CUSTOM,
        CleaningMode.SMART_MODE,
    ]


def test_get_cleaning_mode_parameters_with_brush_up_mop() -> None:
    """Test mop-only uses the brush-up mode when supported."""
    status_trait = _create_cleaning_mode_status_trait(is_support_main_brush_up_down_supported=True)

    assert get_cleaning_mode_parameters(CleaningMode.MOP, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.OFF_RAISE_MAIN_BRUSH.code,
            "water_box_mode": WaterModes.STANDARD.code,
            "mop_mode": CleanRoutes.STANDARD.code,
        }
    ]


def test_get_cleaning_mode_parameters_without_clean_route_setting() -> None:
    """Test older V1 devices use the 2-field clean motor payload."""
    status_trait = _create_cleaning_mode_status_trait(
        is_clean_route_setting_supported=False,
        is_customized_clean_supported=False,
    )

    assert get_cleaning_mode_parameters(CleaningMode.VACUUM, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.BALANCED.code,
            "water_box_mode": WaterModes.OFF.code,
        }
    ]
    assert get_cleaning_mode_parameters(CleaningMode.VAC_AND_MOP, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.BALANCED.code,
            "water_box_mode": WaterModes.STANDARD.code,
        }
    ]
    assert get_cleaning_mode_parameters(CleaningMode.MOP, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.OFF.code,
            "water_box_mode": WaterModes.STANDARD.code,
        }
    ]


def test_get_cleaning_mode_parameters_water_slide_device() -> None:
    """Water-slide devices should use a slide-compatible water code, not 202."""
    status_trait = _create_cleaning_mode_status_trait(is_water_slide_mode_supported=True)

    assert get_cleaning_mode_parameters(CleaningMode.VACUUM, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.BALANCED.code,
            "water_box_mode": WaterModes.OFF.code,
            "mop_mode": CleanRoutes.STANDARD.code,
        }
    ]
    assert get_cleaning_mode_parameters(CleaningMode.VAC_AND_MOP, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.BALANCED.code,
            "water_box_mode": WaterModes.PURE_WATER_FLOW_MIDDLE.code,
            "mop_mode": CleanRoutes.STANDARD.code,
        }
    ]
    assert get_cleaning_mode_parameters(CleaningMode.MOP, status_trait._device_features_trait) == [
        {
            "fan_power": VacuumModes.OFF.code,
            "water_box_mode": WaterModes.PURE_WATER_FLOW_MIDDLE.code,
            "mop_mode": CleanRoutes.STANDARD.code,
        }
    ]


def test_cleaning_mode_options_water_slide_device() -> None:
    """Water-slide devices should not expose unsupported custom or smart water modes."""
    status_trait = _create_cleaning_mode_status_trait(
        is_water_slide_mode_supported=True,
        is_customized_clean_supported=True,
        is_smart_clean_mode_set_supported=True,
    )

    assert status_trait.cleaning_mode_options == [
        CleaningMode.VACUUM,
        CleaningMode.VAC_AND_MOP,
        CleaningMode.MOP,
    ]


def test_current_cleaning_mode_gentle_not_mop_without_pure_mop() -> None:
    """Test code 105 is not treated as mop-only on devices without pure mop."""
    status_trait = _create_cleaning_mode_status_trait(is_pure_clean_mop_supported=False)
    status_trait.fan_power = VacuumModes.GENTLE.code
    status_trait.water_box_mode = WaterModes.STANDARD.code
    status_trait.mop_mode = CleanRoutes.STANDARD.code

    assert status_trait.current_cleaning_mode == CleaningMode.VAC_AND_MOP


def test_water_slide_mode_mapping() -> None:
    """Test feature-aware water mode mapping for water slide mode devices."""
    short_model = mock_data.A114_PRODUCT_DATA["model"].split(".")[-1]
    features = DeviceFeatures.from_feature_flags(
        new_feature_info=int(mock_data.SAROS_10R_DEVICE_DATA["featureSet"]),
        new_feature_info_str=mock_data.SAROS_10R_DEVICE_DATA["newFeatureSet"],
        feature_info=[],
        product_nickname=SHORT_MODEL_TO_ENUM[short_model],
    )
    status_trait = StatusTrait(cast(DeviceFeaturesTrait, features), region="eu")

    assert features.is_water_slide_mode_supported
    assert status_trait.water_mode_mapping == {
        200: "off",
        221: "slight",
        225: "low",
        235: "medium",
        245: "moderate",
        248: "high",
        250: "extreme",
    }
    assert [mode.value for mode in status_trait.water_mode_options] == [
        "off",
        "slight",
        "low",
        "medium",
        "moderate",
        "high",
        "extreme",
    ]

    status_trait.water_box_mode = 225
    assert status_trait.water_mode_name == "low"
    status_trait.water_box_mode = 200
    assert status_trait.water_mode_name == "off"


def test_update_from_dps(status_trait: StatusTrait) -> None:
    """Test updating status from data protocol push message."""
    assert status_trait.battery is None
    assert status_trait.state is None

    status_trait.update_from_dps(
        {
            RoborockDataProtocol.STATE: 5,
            RoborockDataProtocol.BATTERY: 85,
            RoborockDataProtocol.FAN_POWER: 102,
        }
    )

    assert status_trait.state == 5
    assert status_trait.battery == 85
    assert status_trait.fan_power == 102


def test_update_from_dps_partial(status_trait: StatusTrait) -> None:
    """Test that partial updates only modify the specified fields."""
    status_trait.battery = 100
    status_trait.state = RoborockStateCode.charging

    status_trait.update_from_dps(
        {
            RoborockDataProtocol.BATTERY: 90,
        }
    )

    assert status_trait.battery == 90
    assert status_trait.state == RoborockStateCode.charging  # Unchanged


def test_update_listener(status_trait: StatusTrait) -> None:
    """Test that update listeners receive notifications."""
    event = asyncio.Event()
    unsubscribe = status_trait.add_update_listener(event.set)

    status_trait.update_from_dps(
        {
            RoborockDataProtocol.BATTERY: 88,
        }
    )

    assert event.is_set()
    event.clear()

    unsubscribe()

    status_trait.update_from_dps(
        {
            RoborockDataProtocol.BATTERY: 87,
        }
    )

    assert not event.is_set()


def test_update_listener_ignores_unrelated(status_trait: StatusTrait) -> None:
    """Test that update listeners are not notified for unrecognized data points."""
    event = asyncio.Event()
    unsubscribe = status_trait.add_update_listener(event.set)

    # TASK_COMPLETE is not annotated with dps metadata on StatusV2
    status_trait.update_from_dps(
        {
            RoborockDataProtocol.TASK_COMPLETE: 1,
        }
    )

    assert not event.is_set()
    unsubscribe()



"""Roborock exceptions."""

from enum import Enum
from typing import Any


class RoborockException(Exception):
    """Class for Roborock exceptions."""


class RoborockTimeout(RoborockException):
    """Class for Roborock timeout exceptions."""


class RoborockConnectionException(RoborockException):
    """Class for Roborock connection exceptions."""


class RoborockBackoffException(RoborockException):
    """Class for Roborock exceptions when many retries were made."""


class VacuumError(RoborockException):
    """Class for vacuum errors."""


class CommandVacuumError(RoborockException):
    """Class for command vacuum errors."""

    def __init__(self, command: str | None, vacuum_error: VacuumError):
        self.message = f"{command or 'unknown'}: {str(vacuum_error)}"
        super().__init__(self.message)


class UnknownMethodError(RoborockException):
    """Class for an invalid method being sent."""


class RoborockAccountDoesNotExist(RoborockException):
    """Class for Roborock account does not exist exceptions."""


class RoborockUrlException(RoborockException):
    """Class for being unable to get the URL for the Roborock account."""


class RoborockInvalidCode(RoborockException):
    """Class for Roborock invalid code exceptions."""


class RoborockInvalidEmail(RoborockException):
    """Class for Roborock invalid formatted email exceptions."""


class RoborockInvalidUserAgreement(RoborockException):
    """Class for Roborock invalid user agreement exceptions."""


class RoborockNoUserAgreement(RoborockException):
    """Class for Roborock no user agreement exceptions."""


class RoborockInvalidCredentials(RoborockException):
    """Class for Roborock credentials have expired or changed."""


class RoborockTooFrequentCodeRequests(RoborockException):
    """Class for Roborock too frequent code requests exceptions."""


class RoborockMissingParameters(RoborockException):
    """Class for Roborock missing parameters exceptions."""


class RoborockTooManyRequest(RoborockException):
    """Class for Roborock too many request exceptions."""


class RoborockRateLimit(RoborockException):
    """Class for our rate limits exceptions."""


class RoborockNoResponseFromBaseURL(RoborockException):
    """We could not find an url that had a record of the given account."""


class RoborockDeviceBusy(RoborockException):
    """Class for Roborock device busy exceptions."""


class RoborockInvalidStatus(RoborockException):
    """Class for Roborock invalid status exceptions (device action locked)."""


class RoborockUnsupportedFeature(RoborockException):
    """Class for Roborock unsupported feature exceptions."""


class RoborockParsingException(RoborockException):
    """Class for Roborock exceptions when parsing device responses."""

    def __init__(self, trait_name: str, command: Enum | str, payload: Any, inner_error: Exception | str) -> None:
        cmd_name = command.name if isinstance(command, Enum) else str(command)
        self.message = (
            f"Failed to parse {cmd_name} response for {trait_name}. Payload: {payload!r} Error: {inner_error!r}"
        )
        super().__init__(self.message)



"""Traits for Q10 B01 devices."""

from roborock.data.b01_q10.b01_q10_code_mappings import (
    B01_Q10_DP,
    YXCleanType,
    YXDeviceCleanTask,
    YXFanLevel,
)

from .command import CommandTrait


class VacuumTrait:
    """Trait for sending vacuum commands.

    This is a wrapper around the CommandTrait for sending vacuum related
    commands to Q10 devices.
    """

    def __init__(self, command: CommandTrait) -> None:
        """Initialize the VacuumTrait."""
        self._command = command

    async def start_clean(self) -> None:
        """Start a whole-home clean.

        The ``dpStartClean`` (201) command selects a task by code: ``1`` =
        whole-home, ``2`` = segment/room (see :meth:`clean_segments`), ``3`` =
        zone, ``4`` = build map, ``5`` = spot. Whole-home and spot accept the
        bare integer code; segment cleaning needs a room selection (an object
        payload) instead.

        Verified live against ss07 hardware: ``{"dps": {"201": 1}}`` starts a
        whole-home clean (clean_task_type -> 1).
        """
        await self._command.send(command=B01_Q10_DP.START_CLEAN, params=1)

    async def clean_segments(self, segment_ids: list[int]) -> None:
        """Start a room / segment clean for the given segment (room) ids.

        The ids are the same room ids the device reports on its map (see the Q10
        ``MapContentTrait`` -- ``map.rooms``, each with an ``id``).

        Unlike whole-home and spot, ``dpStartClean`` (201) carries the room
        selection as an object: ``{"cmd": <task>, "clean_paramters": [<id>, ...]}``,
        where ``cmd`` is the segment-clean task code.

        Verified live against ss07 hardware: sending
        ``{"dps": {"201": {"cmd": 2, "clean_paramters": [9]}}}`` starts cleaning
        room 9 (clean_task_type -> 2 / electoral).
        """
        await self._command.send(
            command=B01_Q10_DP.START_CLEAN,
            # "clean_paramters" intentionally mirrors the device's misspelling of
            # "parameters" -- the firmware only accepts that exact key.
            params={"cmd": YXDeviceCleanTask.ELECTORAL.code, "clean_paramters": segment_ids},
        )

    async def spot_clean(self) -> None:
        """Start a spot / part clean around the robot's current position.

        Verified live: ``{"dps": {"201": 5}}`` (clean_task_type -> 5).
        """
        await self._command.send(command=B01_Q10_DP.START_CLEAN, params=5)

    async def pause_clean(self) -> None:
        """Pause the current task. Verified live: ``{"dps": {"204": 0}}``."""
        await self._command.send(command=B01_Q10_DP.PAUSE, params=0)

    async def resume_clean(self) -> None:
        """Resume a paused task. Verified live: ``{"dps": {"205": 0}}``."""
        await self._command.send(command=B01_Q10_DP.RESUME, params=0)

    async def stop_clean(self) -> None:
        """Stop / cancel the current task. Verified live: ``{"dps": {"206": 0}}``."""
        await self._command.send(command=B01_Q10_DP.STOP, params=0)

    async def return_to_dock(self) -> None:
        """Send the robot back to the dock to charge.

        Uses ``dpStartBack`` (202) with the back-dock task code ``5`` (charge),
        matching the official app. Verified live: ``{"dps": {"202": 5}}`` puts the
        robot into the returning state. (The other back-dock codes are ``1`` =
        wash mop en route and ``4`` = collect dust en route.)
        """
        await self._command.send(command=B01_Q10_DP.START_BACK, params=5)

    async def empty_dustbin(self) -> None:
        """Empty the dustbin at the dock.

        Verified live: ``{"dps": {"203": 2}}`` triggers dust collection
        (status -> emptying_the_bin). This is a dock task (``dpStartDockTask``),
        distinct from the en-route collect-dust back-dock code.
        """
        await self._command.send(command=B01_Q10_DP.START_DOCK_TASK, params=2)

    async def set_clean_mode(self, mode: YXCleanType) -> None:
        """Set the cleaning mode (vacuum, mop, or both)."""
        await self._command.send(
            command=B01_Q10_DP.CLEAN_MODE,
            params=mode.code,
        )

    async def set_fan_level(self, level: YXFanLevel) -> None:
        """Set the fan suction level."""
        await self._command.send(
            command=B01_Q10_DP.FAN_LEVEL,
            params=level.code,
        )



from __future__ import annotations

import typing
from enum import StrEnum
from typing import TypeVar

from ...exceptions import RoborockUnsupportedFeature
from ..code_mappings import RoborockModeEnum

if typing.TYPE_CHECKING:
    from roborock.device_features import DeviceFeatures


class VacuumModes(RoborockModeEnum):
    GENTLE = ("gentle", 105)
    OFF = ("off", 105)
    QUIET = ("quiet", 101)
    BALANCED = ("balanced", 102)
    TURBO = ("turbo", 103)
    MAX = ("max", 104)
    MAX_PLUS = ("max_plus", 108)
    OFF_RAISE_MAIN_BRUSH = ("off_raise_main_brush", 109)
    CUSTOMIZED = ("custom", 106)
    SMART_MODE = ("smart_mode", 110)


class CleanRoutes(RoborockModeEnum):
    STANDARD = ("standard", 300)
    DEEP = ("deep", 301)
    DEEP_PLUS = ("deep_plus", 303)
    FAST = ("fast", 304)
    DEEP_PLUS_CN = ("deep_plus", 305)
    SMART_MODE = ("smart_mode", 306)
    CUSTOMIZED = ("custom", 302)


class VacuumModesOld(RoborockModeEnum):
    QUIET = ("quiet", 38)
    BALANCED = ("balanced", 60)
    TURBO = ("turbo", 75)
    MAX = ("max", 100)


class WaterModes(RoborockModeEnum):
    OFF = ("off", 200)
    LOW = ("low", 201)
    MILD = ("mild", 201)
    MEDIUM = ("medium", 202)
    STANDARD = ("standard", 202)
    HIGH = ("high", 203)
    INTENSE = ("intense", 203)
    MIN = ("min", 205)
    MAX = ("max", 206)
    CUSTOMIZED = ("custom", 204)
    CUSTOM = ("custom_water_flow", 207)
    EXTREME = ("extreme", 208)
    SMART_MODE = ("smart_mode", 209)
    PURE_WATER_FLOW_START = ("slight", 221)
    PURE_WATER_FLOW_SMALL = ("low", 225)
    PURE_WATER_FLOW_MIDDLE = ("medium", 235)
    PURE_WATER_FLOW_LARGE = ("moderate", 245)
    PURE_WATER_SUPER_BEGIN = ("high", 248)
    PURE_WATER_FLOW_END = ("extreme", 250)


class WashTowelModes(RoborockModeEnum):
    SMART = ("smart", 10)
    LIGHT = ("light", 0)
    BALANCED = ("balanced", 1)
    DEEP = ("deep", 2)
    SUPER_DEEP = ("super_deep", 8)


class CleaningMode(StrEnum):
    """High-level cleaning intent derived from the lower-level motor settings.

    Prefer this abstraction when you want to present or switch between the
    user-facing cleaning behaviors exposed by the app. The lower-level
    `VacuumModes`, `WaterModes`, and `CleanRoutes` enums are still useful for
    fine-grained selection.
    """

    VACUUM = "vacuum"
    VAC_AND_MOP = "vac_and_mop"
    MOP = "mop"
    CUSTOM = "custom"
    SMART_MODE = "smart_mode"


WATER_SLIDE_MODE_MAPPING: dict[int, WaterModes] = {
    200: WaterModes.OFF,
    221: WaterModes.PURE_WATER_FLOW_START,
    225: WaterModes.PURE_WATER_FLOW_SMALL,
    235: WaterModes.PURE_WATER_FLOW_MIDDLE,
    245: WaterModes.PURE_WATER_FLOW_LARGE,
    248: WaterModes.PURE_WATER_SUPER_BEGIN,
    250: WaterModes.PURE_WATER_FLOW_END,
}

ModeEnumT = TypeVar("ModeEnumT", bound=RoborockModeEnum)


def get_wash_towel_modes(features: DeviceFeatures) -> list[WashTowelModes]:
    """Get the valid wash towel modes for the device"""
    modes = [WashTowelModes.LIGHT, WashTowelModes.BALANCED, WashTowelModes.DEEP]
    if features.is_super_deep_wash_supported and not features.is_dirty_replenish_clean_supported:
        modes.append(WashTowelModes.SUPER_DEEP)
    elif features.is_dirty_replenish_clean_supported:
        modes.append(WashTowelModes.SMART)
    return modes


def get_clean_modes(features: DeviceFeatures) -> list[VacuumModes]:
    """Get the valid clean modes for the device - also known as 'fan power' or 'suction mode'"""
    modes = [VacuumModes.QUIET, VacuumModes.BALANCED, VacuumModes.TURBO, VacuumModes.MAX]
    if features.is_max_plus_mode_supported or features.is_none_pure_clean_mop_with_max_plus:
        # If the vacuum has max plus mode supported
        modes.append(VacuumModes.MAX_PLUS)
    if features.is_pure_clean_mop_supported:
        # If the vacuum is capable of 'pure mop clean' aka no vacuum
        if features.is_support_main_brush_up_down_supported:
            modes.append(VacuumModes.OFF_RAISE_MAIN_BRUSH)
        else:
            modes.append(VacuumModes.OFF)
    else:
        # If not, we can add gentle
        modes.append(VacuumModes.GENTLE)
    if features.is_smart_clean_mode_set_supported:
        modes.append(VacuumModes.SMART_MODE)
    if features.is_customized_clean_supported:
        modes.append(VacuumModes.CUSTOMIZED)
    return modes


def get_clean_routes(features: DeviceFeatures, region: str) -> list[CleanRoutes]:
    """The routes that the vacuum will take while mopping"""
    if features.is_none_pure_clean_mop_with_max_plus:
        return [CleanRoutes.FAST, CleanRoutes.STANDARD]
    supported = [CleanRoutes.STANDARD, CleanRoutes.DEEP]
    if features.is_careful_slow_mop_supported:
        if not (
            features.is_corner_clean_mode_supported
            and features.is_clean_route_deep_slow_plus_supported
            and region == "cn"
        ):
            # for some reason there is a china specific deep plus mode
            supported.append(CleanRoutes.DEEP_PLUS_CN)
        else:
            supported.append(CleanRoutes.DEEP_PLUS)

    if features.is_clean_route_fast_mode_supported:
        supported.append(CleanRoutes.FAST)
    if features.is_smart_clean_mode_set_supported:
        supported.append(CleanRoutes.SMART_MODE)
    if features.is_customized_clean_supported:
        supported.append(CleanRoutes.CUSTOMIZED)

    return supported


def get_water_modes(features: DeviceFeatures) -> list[WaterModes]:
    """Get the valid water modes for the device - also known as 'water flow' or 'water level'"""
    # Water slide mode supports a separate set of water flow codes.
    if features.is_water_slide_mode_supported:
        return list(WATER_SLIDE_MODE_MAPPING.values())

    supported_modes = [WaterModes.OFF]
    if features.is_mop_shake_module_supported:
        # For mops that have the vibrating mop pad, they do mild standard intense
        supported_modes.extend([WaterModes.MILD, WaterModes.STANDARD, WaterModes.INTENSE])
    else:
        supported_modes.extend([WaterModes.LOW, WaterModes.MEDIUM, WaterModes.HIGH])
    if features.is_custom_water_box_distance_supported:
        # This is for devices that allow you to set a custom water flow from 0-100
        supported_modes.append(WaterModes.CUSTOM)
    if features.is_mop_shake_module_supported and features.is_mop_shake_water_max_supported:
        supported_modes.append(WaterModes.EXTREME)
    if features.is_smart_clean_mode_set_supported:
        supported_modes.append(WaterModes.SMART_MODE)
    if features.is_customized_clean_supported:
        supported_modes.append(WaterModes.CUSTOMIZED)

    return supported_modes


def get_water_mode_mapping(features: DeviceFeatures) -> dict[int, str]:
    """Get water mode mapping by supported feature set.

    WaterModes contains aliases for multiple codes that share the same value
    string (e.g. low can be 201 or 225). For water slide mode devices we need
    explicit code mapping to preserve those slide-specific codes.
    """
    if features.is_water_slide_mode_supported:
        return {code: mode.value for code, mode in WATER_SLIDE_MODE_MAPPING.items()}
    return {mode.code: mode.value for mode in get_water_modes(features)}


def get_cleaning_mode_options(features: DeviceFeatures) -> list[CleaningMode]:
    """Return the supported high-level cleaning modes for the device.

    These options are the preferred user-facing choices because they bundle the
    correct fan, water, and mop-route settings together for the device. Callers
    should generally present these instead of mixing lower-level mode enums
    unless they explicitly need fine-grained control.
    """
    if not features.is_support_water_mode:
        return []

    supported_water_modes = get_water_modes(features)
    options = [CleaningMode.VACUUM, CleaningMode.VAC_AND_MOP]
    if features.is_pure_clean_mop_supported:
        options.append(CleaningMode.MOP)
    if features.is_customized_clean_supported and WaterModes.CUSTOMIZED in supported_water_modes:
        options.append(CleaningMode.CUSTOM)
    if features.is_smart_clean_mode_set_supported and WaterModes.SMART_MODE in supported_water_modes:
        options.append(CleaningMode.SMART_MODE)
    return options


def get_mop_only_vacuum_mode(features: DeviceFeatures) -> VacuumModes:
    """Determine the vacuum mode to use when you just want to mop.

    There are three cases that must be handled:
    1. The device does not support only mopping.
    2. The device supports raising the vacuum brush while mopping
    3. All other cases.
    """
    if not features.is_pure_clean_mop_supported:
        raise RoborockUnsupportedFeature("Mop-only cleaning is not supported")
    if features.is_support_main_brush_up_down_supported:
        return VacuumModes.OFF_RAISE_MAIN_BRUSH
    return VacuumModes.OFF


def _get_default_mopping_water_mode(features: DeviceFeatures) -> WaterModes:
    """Pick a sensible default water mode when mopping for the device."""
    # Water-slide devices use a disjoint set of water codes; pick a mid-flow
    # slide code instead of the standard 202, which they don't accept.
    if features.is_water_slide_mode_supported:
        return WaterModes.PURE_WATER_FLOW_MIDDLE
    return WaterModes.STANDARD


def _get_clean_motor_mode_params(
    mode: CleaningMode,
    features: DeviceFeatures,
) -> tuple[VacuumModes, WaterModes, CleanRoutes]:
    """Return (fan_power, water_box_mode, mop_mode) enums for the high-level mode."""
    if mode == CleaningMode.VACUUM:
        return (VacuumModes.BALANCED, WaterModes.OFF, CleanRoutes.STANDARD)
    if mode == CleaningMode.VAC_AND_MOP:
        return (VacuumModes.BALANCED, _get_default_mopping_water_mode(features), CleanRoutes.STANDARD)
    if mode == CleaningMode.MOP:
        return (
            get_mop_only_vacuum_mode(features),
            _get_default_mopping_water_mode(features),
            CleanRoutes.STANDARD,
        )
    if mode == CleaningMode.CUSTOM:
        return (VacuumModes.CUSTOMIZED, WaterModes.CUSTOMIZED, CleanRoutes.CUSTOMIZED)
    if mode == CleaningMode.SMART_MODE:
        return (VacuumModes.SMART_MODE, WaterModes.SMART_MODE, CleanRoutes.SMART_MODE)
    raise RoborockUnsupportedFeature(f"Cleaning mode {mode.value!r} is not supported")


def resolve_cleaning_mode(cleaning_mode: str | CleaningMode) -> CleaningMode:
    """Resolve a string or enum into a CleaningMode value."""
    if isinstance(cleaning_mode, CleaningMode):
        return cleaning_mode
    try:
        return CleaningMode(cleaning_mode)
    except ValueError as err:
        raise RoborockUnsupportedFeature(f"Cleaning mode {cleaning_mode!r} is not supported") from err


def get_cleaning_mode_parameters(cleaning_mode: CleaningMode, features: DeviceFeatures) -> list[dict[str, int]]:
    """Get the RPC payload for switching the high-level cleaning mode."""
    if cleaning_mode not in get_cleaning_mode_options(features):
        raise RoborockUnsupportedFeature(f"Cleaning mode {cleaning_mode.value!r} is not supported")

    fan_power, water_box_mode, mop_mode = _get_clean_motor_mode_params(cleaning_mode, features)
    params: dict[str, int] = {"fan_power": fan_power.code, "water_box_mode": water_box_mode.code}
    if features.is_clean_route_setting_supported:
        params["mop_mode"] = mop_mode.code
    return [params]


def _resolve_mode_code(value: int | ModeEnumT | None, mode_cls: type[ModeEnumT]) -> ModeEnumT | None:
    """Resolve a raw code or enum into a RoborockModeEnum."""
    if value is None:
        return None
    if isinstance(value, mode_cls):
        return value
    return mode_cls.from_code_optional(int(value))


def _resolve_clean_mode(value: int | VacuumModes | None, features: DeviceFeatures) -> VacuumModes | None:
    """Resolve a vacuum mode code, accounting for feature-specific code aliases."""
    if value is None or isinstance(value, VacuumModes):
        return value
    if value == VacuumModes.OFF.code:
        if features.is_pure_clean_mop_supported:
            return get_mop_only_vacuum_mode(features)
        return VacuumModes.GENTLE
    return VacuumModes.from_code_optional(value)


def get_current_cleaning_mode(
    clean_mode: int | VacuumModes | None,
    water_mode: int | WaterModes | None,
    mop_mode: int | CleanRoutes | None,
    features: DeviceFeatures,
) -> CleaningMode | None:
    """Classify the current high-level cleaning mode from individual mode codes."""
    if not features.is_support_water_mode:
        return None
    clean_mode_enum = _resolve_clean_mode(clean_mode, features)
    water_mode_enum = _resolve_mode_code(water_mode, WaterModes)
    mop_mode_enum = _resolve_mode_code(mop_mode, CleanRoutes)
    if clean_mode_enum is None or water_mode_enum is None:
        return None

    if is_smart_mode_set(water_mode_enum, clean_mode_enum, mop_mode_enum):
        return CleaningMode.SMART_MODE
    if is_mode_customized(clean_mode_enum, water_mode_enum, mop_mode_enum):
        return CleaningMode.CUSTOM
    if water_mode_enum != WaterModes.OFF:
        try:
            if clean_mode_enum == get_mop_only_vacuum_mode(features):
                return CleaningMode.MOP
        except RoborockUnsupportedFeature:
            pass
    if water_mode_enum == WaterModes.OFF:
        return CleaningMode.VACUUM
    return CleaningMode.VAC_AND_MOP


def is_mode_customized(
    clean_mode: VacuumModes | None,
    water_mode: WaterModes | None,
    mop_mode: CleanRoutes | None,
) -> bool:
    """Check if any of the cleaning modes are set to a custom value."""
    return (
        clean_mode == VacuumModes.CUSTOMIZED
        or water_mode == WaterModes.CUSTOMIZED
        or mop_mode == CleanRoutes.CUSTOMIZED
    )


def is_smart_mode_set(
    water_mode: WaterModes | None,
    clean_mode: VacuumModes | None,
    mop_mode: CleanRoutes | None,
) -> bool:
    """Check if the smart mode is set for the given water mode and clean mode"""
    return (
        water_mode == WaterModes.SMART_MODE
        or clean_mode == VacuumModes.SMART_MODE
        or mop_mode == CleanRoutes.SMART_MODE
    )



from collections.abc import Awaitable, Callable
from typing import Any

import pytest

from roborock.data.b01_q10.b01_q10_code_mappings import YXCleanType, YXFanLevel
from roborock.devices.traits.b01.q10 import Q10PropertiesApi
from roborock.devices.traits.b01.q10.vacuum import VacuumTrait

from .conftest import FakeB01Q10Channel


@pytest.fixture(name="vacuum")
def vacuum_fixture(q10_api: Q10PropertiesApi) -> VacuumTrait:
    return q10_api.vacuum


@pytest.mark.parametrize(
    ("command_fn", "expected_payload"),
    [
        # Payloads verified live against ss07 hardware.
        (lambda x: x.start_clean(), {"201": 1}),
        (lambda x: x.clean_segments([9]), {"201": {"cmd": 2, "clean_paramters": [9]}}),
        (lambda x: x.clean_segments([1, 2]), {"201": {"cmd": 2, "clean_paramters": [1, 2]}}),
        (lambda x: x.spot_clean(), {"201": 5}),
        (lambda x: x.pause_clean(), {"204": 0}),
        (lambda x: x.resume_clean(), {"205": 0}),
        (lambda x: x.stop_clean(), {"206": 0}),
        (lambda x: x.return_to_dock(), {"202": 5}),
        (lambda x: x.empty_dustbin(), {"203": 2}),
        (lambda x: x.set_clean_mode(YXCleanType.VAC_AND_MOP), {"137": 1}),
        (lambda x: x.set_fan_level(YXFanLevel.BALANCED), {"123": 2}),
    ],
)
async def test_vacuum_commands(
    vacuum: VacuumTrait,
    fake_channel: FakeB01Q10Channel,
    command_fn: Callable[[VacuumTrait], Awaitable[None]],
    expected_payload: dict[str, Any],
) -> None:
    """Test sending a vacuum start command."""
    await command_fn(vacuum)

    assert len(fake_channel.published_commands) == 1
    command, params = fake_channel.published_commands[0]

    dp_code = int(list(expected_payload.keys())[0])
    expected_params = list(expected_payload.values())[0]

    assert command.code == dp_code
    assert params == expected_params



"""Tests for the callbacks module."""

import logging
from unittest.mock import Mock

from roborock.callbacks import CallbackList, CallbackMap, safe_callback


def test_safe_callback_successful_execution():
    """Test that safe_callback executes callback successfully."""
    mock_callback = Mock()
    wrapped = safe_callback(mock_callback)

    wrapped("test_value")

    mock_callback.assert_called_once_with("test_value")


def test_safe_callback_catches_exception():
    """Test that safe_callback catches and logs exceptions."""

    def failing_callback(value):
        raise ValueError("Test exception")

    mock_logger = Mock(spec=logging.Logger)
    wrapped = safe_callback(failing_callback, mock_logger)

    # Should not raise exception
    wrapped("test_value")

    mock_logger.error.assert_called_once()
    assert "Uncaught error in callback" in mock_logger.error.call_args[0][0]


def test_safe_callback_uses_default_logger():
    """Test that safe_callback uses default logger when none provided."""

    def failing_callback(value):
        raise ValueError("Test exception")

    wrapped = safe_callback(failing_callback)

    # Should not raise exception
    wrapped("test_value")


# CallbackMap tests


def test_callback_map_add_callback_and_invoke():
    """Test adding callback and invoking it."""
    callback_map = CallbackMap[str, str]()
    mock_callback = Mock()

    remove_fn = callback_map.add_callback("key1", mock_callback)
    callback_map("key1", "test_value")

    mock_callback.assert_called_once_with("test_value")
    assert callable(remove_fn)


def test_callback_map_multiple_callbacks_same_key():
    """Test multiple callbacks for the same key."""
    callback_map = CallbackMap[str, str]()
    mock_callback1 = Mock()
    mock_callback2 = Mock()

    callback_map.add_callback("key1", mock_callback1)
    callback_map.add_callback("key1", mock_callback2)
    callback_map("key1", "test_value")

    mock_callback1.assert_called_once_with("test_value")
    mock_callback2.assert_called_once_with("test_value")


def test_callback_map_different_keys():
    """Test callbacks for different keys."""
    callback_map = CallbackMap[str, str]()
    mock_callback1 = Mock()
    mock_callback2 = Mock()

    callback_map.add_callback("key1", mock_callback1)
    callback_map.add_callback("key2", mock_callback2)

    callback_map("key1", "value1")
    callback_map("key2", "value2")

    mock_callback1.assert_called_once_with("value1")
    mock_callback2.assert_called_once_with("value2")


def test_callback_map_get_callbacks():
    """Test getting callbacks for a key."""
    callback_map = CallbackMap[str, str]()
    mock_callback = Mock()

    # No callbacks initially
    assert callback_map.get_callbacks("key1") == []

    # Add callback
    callback_map.add_callback("key1", mock_callback)
    callbacks = callback_map.get_callbacks("key1")

    assert len(callbacks) == 1
    assert callbacks[0] == mock_callback


def test_callback_map_remove_callback():
    """Test removing callback."""
    callback_map = CallbackMap[str, str]()
    mock_callback = Mock()

    remove_fn = callback_map.add_callback("key1", mock_callback)

    # Callback should be there
    assert len(callback_map.get_callbacks("key1")) == 1

    # Remove callback
    remove_fn()

    # Callback should be gone
    assert callback_map.get_callbacks("key1") == []


def test_callback_map_remove_callback_cleans_up_key():
    """Test that removing last callback for a key removes the key."""
    callback_map = CallbackMap[str, str]()
    mock_callback = Mock()

    remove_fn = callback_map.add_callback("key1", mock_callback)

    # Key should exist
    assert "key1" in callback_map._callbacks

    # Remove callback
    remove_fn()

    # Key should be removed
    assert "key1" not in callback_map._callbacks


def test_callback_map_exception_handling(caplog):
    """Test that exceptions in callbacks are handled gracefully."""
    callback_map = CallbackMap[str, str]()

    def failing_callback(value):
        raise ValueError("Test exception")

    callback_map.add_callback("key1", failing_callback)

    with caplog.at_level(logging.ERROR):
        callback_map("key1", "test_value")

    assert "Uncaught error in callback" in caplog.text


def test_callback_map_custom_logger():
    """Test using custom logger."""
    mock_logger = Mock(spec=logging.Logger)
    callback_map = CallbackMap[str, str](logger=mock_logger)

    def failing_callback(value):
        raise ValueError("Test exception")

    callback_map.add_callback("key1", failing_callback)
    callback_map("key1", "test_value")

    mock_logger.error.assert_called_once()


# CallbackList tests


def test_callback_list_add_callback_and_invoke():
    """Test adding callback and invoking it."""
    callback_list = CallbackList[str]()
    mock_callback = Mock()

    remove_fn = callback_list.add_callback(mock_callback)
    callback_list("test_value")

    mock_callback.assert_called_once_with("test_value")
    assert callable(remove_fn)


def test_callback_list_multiple_callbacks():
    """Test multiple callbacks in the list."""
    callback_list = CallbackList[str]()
    mock_callback1 = Mock()
    mock_callback2 = Mock()

    callback_list.add_callback(mock_callback1)
    callback_list.add_callback(mock_callback2)
    callback_list("test_value")

    mock_callback1.assert_called_once_with("test_value")
    mock_callback2.assert_called_once_with("test_value")


def test_callback_list_remove_callback():
    """Test removing callback from list."""
    callback_list = CallbackList[str]()
    mock_callback1 = Mock()
    mock_callback2 = Mock()

    remove_fn1 = callback_list.add_callback(mock_callback1)
    callback_list.add_callback(mock_callback2)

    # Both should be called
    callback_list("test_value")
    assert mock_callback1.call_count == 1
    assert mock_callback2.call_count == 1

    # Remove first callback
    remove_fn1()

    # Only second should be called
    callback_list("test_value2")
    assert mock_callback1.call_count == 1  # Still 1
    assert mock_callback2.call_count == 2  # Now 2


def test_callback_list_exception_handling(caplog):
    """Test that exceptions in callbacks are handled gracefully."""
    callback_list = CallbackList[str]()

    def failing_callback(value):
        raise ValueError("Test exception")

    callback_list.add_callback(failing_callback)

    with caplog.at_level(logging.ERROR):
        callback_list("test_value")

    assert "Uncaught error in callback" in caplog.text


def test_callback_list_custom_logger():
    """Test using custom logger."""
    mock_logger = Mock(spec=logging.Logger)
    callback_list = CallbackList[str](logger=mock_logger)

    def failing_callback(value):
        raise ValueError("Test exception")

    callback_list.add_callback(failing_callback)
    callback_list("test_value")

    mock_logger.error.assert_called_once()



"""Common code for MQTT tests."""

import asyncio
import datetime
from collections.abc import AsyncGenerator, Callable, Generator
from queue import Queue
from typing import Any
from unittest.mock import patch

import paho.mqtt.client as mqtt
import pytest

from .mqtt import FakeMqttSocketHandler


@pytest.fixture(name="mock_aiomqtt_client")
async def mock_aiomqtt_client_fixture() -> AsyncGenerator[None, None]:
    """Fixture to patch the MQTT underlying sync client.

    The tests use fake sockets, so this ensures that the async mqtt client does not
    attempt to listen on them directly. We instead just poll the socket for
    data ourselves.
    """

    event_loop = asyncio.get_running_loop()

    orig_class = mqtt.Client

    async def poll_sockets(client: mqtt.Client) -> None:
        """Poll the mqtt client sockets in a loop to pick up new data."""
        try:
            while True:
                event_loop.call_soon_threadsafe(client.loop_read)
                event_loop.call_soon_threadsafe(client.loop_write)
                await asyncio.sleep(0.01)
        except asyncio.CancelledError:
            pass

    task: asyncio.Task[None] | None = None

    def new_client(*args: Any, **kwargs: Any) -> mqtt.Client:
        """Create a new mqtt client and start the socket polling task."""
        nonlocal task
        client = orig_class(*args, **kwargs)
        task = event_loop.create_task(poll_sockets(client))
        return client

    with (
        patch("aiomqtt.client.Client._on_socket_open"),
        patch("aiomqtt.client.Client._on_socket_close"),
        patch("aiomqtt.client.Client._on_socket_register_write"),
        patch("aiomqtt.client.Client._on_socket_unregister_write"),
        patch("aiomqtt.client.mqtt.Client", side_effect=new_client),
    ):
        yield
        if task:
            task.cancel()
            await task


@pytest.fixture
def fast_backoff_fixture() -> Generator[None, None, None]:
    """Fixture to speed up backoff."""
    with patch(
        "roborock.mqtt.roborock_session.MIN_BACKOFF_INTERVAL",
        datetime.timedelta(seconds=0.01),
    ):
        yield


@pytest.fixture
def push_mqtt_response(
    mqtt_response_queue: Queue, fake_mqtt_socket_handler: FakeMqttSocketHandler
) -> Callable[[bytes], None]:
    """Fixture to push a response to the client."""

    def _push(data: bytes) -> None:
        mqtt_response_queue.put(data)
        fake_mqtt_socket_handler.push_response()

    return _push



"""Tests for the DeviceFeaturesTrait related functionality."""

import pytest
from syrupy import SnapshotAssertion

from roborock.data import HomeDataDevice
from roborock.data.v1.v1_containers import ConsumableField, StatusField
from roborock.devices.device import RoborockDevice
from roborock.devices.traits.v1.consumeable import ConsumableTrait
from roborock.devices.traits.v1.status import StatusTrait
from tests import mock_data

V1_DEVICES = {
    k: HomeDataDevice.from_dict(device) for k, device in mock_data.DEVICES.items() if device.get("pv") == "1.0"
}


@pytest.mark.parametrize(
    ("device_info"),
    V1_DEVICES.values(),
    ids=list(V1_DEVICES.keys()),
)
async def test_is_attribute_supported(
    device_info: HomeDataDevice,
    device: RoborockDevice,
    snapshot: SnapshotAssertion,
) -> None:
    """Test if a field is supported."""
    assert device.v1_properties is not None
    assert device.v1_properties.device_features is not None
    device_features_trait = device.v1_properties.device_features

    is_v1_supported = {
        field.value: device_features_trait.is_field_supported(StatusTrait, field) for field in StatusField
    }
    assert is_v1_supported == snapshot


@pytest.mark.parametrize(
    ("device_info"),
    V1_DEVICES.values(),
    ids=list(V1_DEVICES.keys()),
)
async def test_is_consumable_field_supported(
    device_info: HomeDataDevice,
    device: RoborockDevice,
    snapshot: SnapshotAssertion,
) -> None:
    """Test if a field is supported."""
    assert device.v1_properties is not None
    assert device.v1_properties.device_features is not None
    device_features_trait = device.v1_properties.device_features

    is_v1_supported = {
        field.value: device_features_trait.is_field_supported(ConsumableTrait, field) for field in ConsumableField
    }
    assert is_v1_supported == snapshot



"""Tests for the a01_channel."""

from typing import Any

import pytest

from roborock.devices.rpc.a01_channel import send_decoded_command
from roborock.protocols.a01_protocol import encode_mqtt_payload
from roborock.roborock_message import (
    RoborockDyadDataProtocol,
    RoborockMessage,
    RoborockMessageProtocol,
)
from tests.fixtures.channel_fixtures import FakeChannel


@pytest.fixture
def mock_mqtt_channel() -> FakeChannel:
    """Fixture for a fake MQTT channel."""
    return FakeChannel()


async def test_id_query(mock_mqtt_channel: FakeChannel):
    """Test successful command sending and response decoding."""
    # Command parameters to send
    params: dict[RoborockDyadDataProtocol, Any] = {
        RoborockDyadDataProtocol.ID_QUERY: [
            RoborockDyadDataProtocol.WARM_LEVEL,
            RoborockDyadDataProtocol.POWER,
        ]
    }
    encoded = encode_mqtt_payload(
        {
            RoborockDyadDataProtocol.WARM_LEVEL: 101,
            RoborockDyadDataProtocol.POWER: 75,
        },
        value_encoder=lambda x: x,
    )
    response_message = RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_RESPONSE, payload=encoded.payload, version=encoded.version
    )
    mock_mqtt_channel.response_queue.append(response_message)

    # Call the function to be tested
    result = await send_decoded_command(mock_mqtt_channel, params)  # type: ignore[call-overload]

    # Assertions
    assert result == {
        RoborockDyadDataProtocol.WARM_LEVEL: 101,
        RoborockDyadDataProtocol.POWER: 75,
    }
    mock_mqtt_channel.publish.assert_awaited_once()
    mock_mqtt_channel.subscribe.assert_awaited_once()



"""Trait for managing consumable attributes.

A consumable attribute is one that is expected to be replaced or refilled
periodically, such as filters, brushes, etc.
"""

import logging
from enum import StrEnum
from typing import Any, Self

from roborock.data import Consumable
from roborock.devices.traits.common import DpsDataConverter, TraitUpdateListener
from roborock.devices.traits.v1 import common
from roborock.roborock_message import RoborockDataProtocol
from roborock.roborock_typing import RoborockCommand

__all__ = [
    "ConsumableTrait",
]

_LOGGER = logging.getLogger(__name__)

_DPS_CONVERTER = DpsDataConverter.from_dataclass(Consumable)


class ConsumableAttribute(StrEnum):
    """Enum for consumable attributes."""

    SENSOR_DIRTY_TIME = "sensor_dirty_time"
    FILTER_WORK_TIME = "filter_work_time"
    SIDE_BRUSH_WORK_TIME = "side_brush_work_time"
    MAIN_BRUSH_WORK_TIME = "main_brush_work_time"
    STRAINER_WORK_TIME = "strainer_work_time"
    CLEANING_BRUSH_WORK_TIME = "cleaning_brush_work_time"

    @classmethod
    def from_str(cls, value: str) -> Self:
        """Create a ConsumableAttribute from a string value."""
        for member in cls:
            if member.value == value:
                return member
        raise ValueError(f"Unknown ConsumableAttribute: {value}")


class ConsumableTrait(Consumable, common.V1TraitMixin, TraitUpdateListener):
    """Trait for managing consumable attributes on Roborock devices.

    After the first refresh, you can tell what consumables are supported by
    checking which attributes are not None.
    """

    command = RoborockCommand.GET_CONSUMABLE
    converter = common.DefaultConverter(Consumable)

    def __init__(self) -> None:
        """Initialize the consumable trait."""
        super().__init__()
        TraitUpdateListener.__init__(self, logger=_LOGGER)

    async def reset_consumable(self, consumable: ConsumableAttribute) -> None:
        """Reset a specific consumable attribute on the device."""
        await self.rpc_channel.send_command(RoborockCommand.RESET_CONSUMABLE, params=[consumable.value])
        await self.refresh()

    def update_from_dps(self, decoded_dps: dict[RoborockDataProtocol, Any]) -> None:
        """Update the trait from data protocol push message data.

        This handles unsolicited status updates pushed by the device
        via RoborockDataProtocol codes (e.g. STATE=121, BATTERY=122).
        """
        if _DPS_CONVERTER.update_from_dps(self, decoded_dps):
            self._notify_update()



from collections.abc import Awaitable, Callable
from typing import Any

import pytest

from roborock.devices.traits.b01.q10 import Q10PropertiesApi
from roborock.devices.traits.b01.q10.remote import RemoteTrait

from .conftest import FakeB01Q10Channel


@pytest.fixture(name="remote")
def remote_fixture(q10_api: Q10PropertiesApi) -> RemoteTrait:
    return q10_api.remote


@pytest.mark.parametrize(
    ("command_fn", "expected_payload"),
    [
        (lambda x: x.forward(), {"101": {"12": 0}}),
        (lambda x: x.left(), {"101": {"12": 2}}),
        (lambda x: x.right(), {"101": {"12": 3}}),
        (lambda x: x.stop(), {"101": {"12": 4}}),
        (lambda x: x.exit_remote(), {"101": {"12": 5}}),
    ],
)
async def test_remote_commands(
    remote: RemoteTrait,
    fake_channel: FakeB01Q10Channel,
    command_fn: Callable[[RemoteTrait], Awaitable[None]],
    expected_payload: dict[str, Any],
) -> None:
    """Test sending a remote start command."""
    await command_fn(remote)

    assert len(fake_channel.published_commands) == 1
    command, params = fake_channel.published_commands[0]

    dp_code = int(list(expected_payload.keys())[0])
    expected_params = list(expected_payload.values())[0]

    assert command.code == dp_code
    assert params == expected_params



"""End-to-end tests package."""

pytest_plugins = [
    "tests.fixtures.logging_fixtures",
    "tests.fixtures.local_async_fixtures",
    "tests.fixtures.pahomqtt_fixtures",
    "tests.fixtures.aiomqtt_fixtures",
    "tests.fixtures.web_api_fixtures",
]



from roborock.broadcast_protocol import RoborockProtocol


def test_l01_data():
    data = bytes.fromhex(
        "4c30310000000000000043841496d5a31e34b5b02c1867c445509ba5a21aec1fa4b307bddeb27a75d9b366193e8a97d0534dc39851c"
        "980609f2670cdcaee04594ec5c93e3c5ae609b0c9a203139ac8e40c8c"
    )
    prot = RoborockProtocol()
    prot.datagram_received(data, None)
    device = prot.devices_found[0]
    assert device.duid == "ZrQn1jfZtJQLoPOL7620e"
    assert device.ip == "192.168.1.4"
    assert device.version == b"L01"


def test_v1_data():
    data = bytes.fromhex(
        "312e30000003e003e80040b87035058b439f36af42f249605f8661897173f111bb849a6231831f5874a0cf220a25872ea412d796b4902ee"
        "57fdc120074b901b482acb1fe6d06317e3a72ddac654fe0"
    )
    prot = RoborockProtocol()
    prot.datagram_received(data, None)
    device = prot.devices_found[0]
    assert device.duid == "h96rOV3e8DTPMAOLiypREl"
    assert device.ip == "192.168.20.250"
    assert device.version == b"1.0"



"""Fixtures for V1 trait tests."""

from copy import deepcopy
from unittest.mock import AsyncMock, Mock

import pytest

from roborock.data import HomeData, HomeDataDevice, HomeDataProduct, RoborockDockTypeCode, S7MaxVStatus, UserData
from roborock.devices.cache import Cache, DeviceCache, InMemoryCache
from roborock.devices.device import RoborockDevice
from roborock.devices.traits import v1
from tests import mock_data

USER_DATA = UserData.from_dict(mock_data.USER_DATA)
HOME_DATA = HomeData.from_dict(mock_data.HOME_DATA_RAW)
STATUS = S7MaxVStatus.from_dict(mock_data.STATUS)


@pytest.fixture(autouse=True, name="channel")
def device_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="mock_rpc_channel")
def rpc_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="mock_mqtt_rpc_channel")
def mqtt_rpc_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="mock_map_rpc_channel")
def map_rpc_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="web_api_client")
def web_api_client_fixture() -> AsyncMock:
    """Fixture to set up the web API client for tests."""
    return AsyncMock()


@pytest.fixture(name="trait_home_data")
def trait_home_data_fixture(request: pytest.FixtureRequest) -> HomeData:
    """Fixture to provide HomeData, optionally overridden via indirect parametrization."""
    return deepcopy(getattr(request, "param", HOME_DATA))


@pytest.fixture(autouse=True, name="roborock_cache")
def roborock_cache_fixture() -> Cache:
    """Fixture to provide a NoCache instance for tests."""
    return InMemoryCache()


@pytest.fixture(autouse=True, name="device_cache")
def device_cache_fixture(roborock_cache: Cache, trait_home_data: HomeData) -> DeviceCache:
    """Fixture to provide a DeviceCache instance for tests."""
    return DeviceCache(trait_home_data.get_all_devices()[0].duid, roborock_cache)


@pytest.fixture(name="device_info")
def device_info_fixture(trait_home_data: HomeData) -> HomeDataDevice:
    """Fixture to provide a DeviceInfo instance for tests."""
    return trait_home_data.get_all_devices()[0]


@pytest.fixture(name="products")
def products_fixture() -> list[HomeDataProduct]:
    """Fixture to provide a Product instance for tests."""
    return [HomeDataProduct.from_dict(product) for product in mock_data.PRODUCTS.values()]


@pytest.fixture(autouse=True, name="device")
def device_fixture(
    channel: AsyncMock,
    mock_rpc_channel: AsyncMock,
    mock_mqtt_rpc_channel: AsyncMock,
    mock_map_rpc_channel: AsyncMock,
    web_api_client: AsyncMock,
    device_cache: DeviceCache,
    device_info: HomeDataDevice,
    trait_home_data: HomeData,
    products: list[HomeDataProduct],
) -> RoborockDevice:
    """Fixture to set up the device for tests."""
    product = next(filter(lambda product: product.id == device_info.product_id, products))
    return RoborockDevice(
        device_info=device_info,
        product=product,
        channel=channel,
        trait=v1.create(
            device_info.duid,
            product,
            trait_home_data,
            mock_rpc_channel,
            mock_mqtt_rpc_channel,
            mock_map_rpc_channel,
            Mock(),
            web_api_client,
            device_cache=device_cache,
            region=USER_DATA.region,
        ),
    )


@pytest.fixture(name="dock_type_code", autouse=True)
def dock_type_code_fixture(request: pytest.FixtureRequest) -> RoborockDockTypeCode | None:
    """Fixture to provide the dock type code for parameterized tests."""
    return RoborockDockTypeCode.s7_max_ultra_dock


@pytest.fixture(autouse=True)
async def discover_features_fixture(
    device: RoborockDevice,
    mock_rpc_channel: AsyncMock,
    dock_type_code: RoborockDockTypeCode | None,
) -> None:
    """Fixture to handle device feature discovery."""
    assert device.v1_properties
    mock_rpc_channel.send_command.side_effect = [
        [mock_data.APP_GET_INIT_STATUS],
        {
            **mock_data.STATUS,
            "dock_type": dock_type_code,
        },
    ]
    # Connecting triggers device discovery
    await device.connect()
    assert device.v1_properties.status.dock_type == dock_type_code
    mock_rpc_channel.send_command.reset_mock()
    mock_rpc_channel.send_command.side_effect = None



"""Thin wrapper around the MQTT channel for Roborock B01 Q7 devices."""

import asyncio
import json
import logging
from collections.abc import Callable
from typing import Protocol, TypeAlias, TypeVar

from roborock.data import HomeDataDevice, HomeDataProduct
from roborock.devices.transport.channel import Channel
from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.exceptions import RoborockException
from roborock.protocols.b01_q7_protocol import (
    B01_Q7_DPS,
    B01_VERSION,
    CommandType,
    MapKey,
    ParamsType,
    Q7RequestMessage,
    create_map_key,
    decode_map_payload,
    decode_rpc_response,
    encode_mqtt_payload,
)
from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol

_LOGGER = logging.getLogger(__name__)
_TIMEOUT = 10.0
_T = TypeVar("_T")
DecodedB01Response: TypeAlias = dict[str, object] | str


class Q7RpcChannel(Protocol):
    """Protocol for Q7 RPC channels."""

    async def send_command(
        self,
        command: CommandType,
        params: ParamsType = None,
    ) -> DecodedB01Response:
        """Send a command and get a decoded response."""
        ...


class Q7MapRpcChannel(Protocol):
    """Protocol for Q7 map RPC channels."""

    async def send_map_command(
        self,
        command: CommandType,
        params: ParamsType = None,
    ) -> bytes:
        """Send a map command and get decoded bytes."""
        ...


def _matches_map_response(response_message: RoborockMessage, *, version: bytes | None) -> bytes | None:
    """Return raw map payload bytes for matching MAP_RESPONSE messages."""
    if (
        response_message.protocol == RoborockMessageProtocol.MAP_RESPONSE
        and response_message.payload
        and response_message.version == version
    ):
        return response_message.payload
    return None


async def _send_command(
    mqtt_channel: MqttChannel,
    request_message: Q7RequestMessage,
    *,
    response_matcher: Callable[[RoborockMessage], _T | None],
) -> _T:
    """Publish a B01 command and resolve on the first matching response."""
    roborock_message = encode_mqtt_payload(request_message)
    future: asyncio.Future[_T] = asyncio.get_running_loop().create_future()

    def on_message(response_message: RoborockMessage) -> None:
        if future.done():
            return
        try:
            response = response_matcher(response_message)
        except Exception as ex:
            future.set_exception(ex)
            return
        if response is not None:
            future.set_result(response)

    unsub = await mqtt_channel.subscribe(on_message)
    try:
        await mqtt_channel.publish(roborock_message)
        return await asyncio.wait_for(future, timeout=_TIMEOUT)
    finally:
        unsub()


async def send_decoded_command(
    mqtt_channel: MqttChannel,
    request_message: Q7RequestMessage,
) -> DecodedB01Response:
    """Send a command on the MQTT channel and get a decoded response."""
    _LOGGER.debug("Sending B01 MQTT command: %s", request_message)

    def find_response(response_message: RoborockMessage) -> DecodedB01Response | None:
        """Handle incoming messages and resolve the future."""
        try:
            decoded_dps = decode_rpc_response(response_message)
        except RoborockException as ex:
            _LOGGER.debug(
                "Failed to decode B01 RPC response (expecting method=%s msg_id=%s): %s: %s",
                request_message.command,
                request_message.msg_id,
                response_message,
                ex,
            )
            return None
        for dps_value in decoded_dps.values():
            # valid responses are JSON strings wrapped in the dps value
            if not isinstance(dps_value, str):
                _LOGGER.debug("Received unexpected response: %s", dps_value)
                continue

            try:
                inner = json.loads(dps_value)
            except (json.JSONDecodeError, TypeError):
                _LOGGER.debug("Received unexpected response: %s", dps_value)
                continue
            if isinstance(inner, dict) and inner.get("msgId") == str(request_message.msg_id):
                _LOGGER.debug("Received query response: %s", inner)
                code = inner.get("code", 0)
                if code != 0:
                    error_msg = f"B01 command failed with code {code} ({request_message})"
                    _LOGGER.debug("B01 error response: %s", error_msg)
                    raise RoborockException(error_msg)
                data = inner.get("data")
                if request_message.command == "prop.get" and not isinstance(data, dict):
                    raise RoborockException(f"Unexpected data type for response {data} ({request_message})")
                return data
        return None

    try:
        return await _send_command(
            mqtt_channel,
            request_message,
            response_matcher=find_response,
        )
    except TimeoutError as ex:
        raise RoborockException(f"B01 command timed out after {_TIMEOUT}s ({request_message})") from ex
    except RoborockException as ex:
        _LOGGER.warning(
            "Error sending B01 decoded command (%s): %s",
            request_message,
            ex,
        )
        raise
    except Exception as ex:
        _LOGGER.exception(
            "Error sending B01 decoded command (%s): %s",
            request_message,
            ex,
        )
        raise


class B01Q7Channel(Channel, Q7RpcChannel, Q7MapRpcChannel):
    """Unified B01 Q7 channel wrapping MQTT transport."""

    def __init__(self, mqtt_channel: MqttChannel, map_key: MapKey) -> None:
        self._mqtt_channel = mqtt_channel
        self._map_key = map_key

    @property
    def is_connected(self) -> bool:
        return self._mqtt_channel.is_connected

    @property
    def is_local_connected(self) -> bool:
        return False

    async def subscribe(self, callback: Callable[[RoborockMessage], None]) -> Callable[[], None]:
        return await self._mqtt_channel.subscribe(callback)

    async def send_command(
        self,
        command: CommandType,
        params: ParamsType = None,
    ) -> DecodedB01Response:
        return await send_decoded_command(
            self._mqtt_channel,
            Q7RequestMessage(dps=B01_Q7_DPS, command=command, params=params),
        )

    async def send_map_command(
        self,
        command: CommandType,
        params: ParamsType = None,
    ) -> bytes:
        """Send a map upload command and return decoded SCMap bytes."""
        request_message = Q7RequestMessage(dps=B01_Q7_DPS, command=command, params=params)
        try:
            raw_payload = await _send_command(
                self._mqtt_channel,
                request_message,
                response_matcher=lambda response_message: _matches_map_response(response_message, version=B01_VERSION),
            )
        except TimeoutError as ex:
            raise RoborockException(f"B01 map command timed out after {_TIMEOUT}s ({request_message})") from ex

        return decode_map_payload(raw_payload, map_key=self._map_key)


def create_b01_q7_channel(
    device: HomeDataDevice,
    product: HomeDataProduct,
    mqtt_channel: MqttChannel,
) -> B01Q7Channel:
    """Create a B01Q7Channel for the given device."""
    if device.sn is None or product.model is None:
        raise RoborockException(
            f"Device serial number and product model are required (sn: {device.sn}, model: {product.model})"
        )
    map_key = create_map_key(serial=device.sn, model=product.model)
    return B01Q7Channel(mqtt_channel, map_key)
